/* Copyright (C) 2007 Christian Dywan <christian at twotoasts dot de>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. */

#include "midori.h"

/*
GList* gtk_window_list_toplevels(void)
void gtk_window_present_with_time(GtkWindow*, guint32 timestamp)
*/

// -- Global variables begin here

GList* browser_list; // Items of type 'CBrowser'
static int browser_count = 0;
CConfig* config;
GtkAccelGroup* accel_group;
GBookmarkFile* bookmarks;
GtkWidget* settings_dialog;
GtkWidget* color_picker;
GtkTooltips* tooltips;

// -- Helper functions begin here

GtkIconTheme* get_icon_theme(GtkWidget* widget)
{
    return gtk_icon_theme_get_for_screen(gtk_widget_get_screen(widget));
}

GtkWidget* menu_item_new(const gchar* sText, const gchar* sIcon
 , const gchar* sPath, const gchar* sKey, GdkModifierType accel_mods
 , GCallback signal, gboolean bSensitive, gpointer userdata)
{
    DEBUG_FUNC
    GtkWidget* menuitem;
    if(sText)
    {
     menuitem = gtk_image_menu_item_new_with_mnemonic(sText);
     if(sIcon)
     {
      GtkWidget* icon = gtk_image_new_from_stock(sIcon, GTK_ICON_SIZE_MENU);
      if(gtk_image_get_storage_type(GTK_IMAGE(icon)) == GTK_IMAGE_EMPTY)
       icon = gtk_image_new_from_icon_name(sIcon, GTK_ICON_SIZE_MENU);
      if(gtk_image_get_storage_type(GTK_IMAGE(icon)) != GTK_IMAGE_EMPTY)
       gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(menuitem), icon);
      else
       g_print("Note: The icon %s is not available.", sIcon);
     }
    }
    else
     menuitem = gtk_image_menu_item_new_from_stock(sIcon, NULL);
    if(signal)
     g_signal_connect(menuitem, "activate", GTK_SIGNAL_FUNC(signal), userdata);
    gtk_widget_set_sensitive(GTK_WIDGET(menuitem), bSensitive && signal);
    sokoke_menu_item_set_accel(GTK_MENU_ITEM(menuitem), sPath, sKey, accel_mods);
    return menuitem;
}

GtkToolItem* tool_button_new(const gchar* sText, const gchar* sIcon
 , gboolean bImportant, gboolean bSensitive, GCallback signal
 , const gchar* sTooltip, gpointer userdata)
{
    DEBUG_FUNC
    GtkToolItem* toolitem;
    if(sText)
    {
     toolitem = gtk_tool_button_new(NULL, sText);
     GtkWidget* icon = gtk_image_new_from_icon_name(sIcon, GTK_ICON_SIZE_MENU);
     GtkIconTheme* icon_theme = get_icon_theme(GTK_WIDGET(toolitem));
     if(gtk_icon_theme_has_icon(icon_theme, sIcon))
      gtk_tool_button_set_icon_name(GTK_TOOL_BUTTON(toolitem), sIcon);
    }
    else
     toolitem = gtk_tool_button_new_from_stock(sIcon);
    if(bImportant)
     gtk_tool_item_set_is_important(toolitem, TRUE);
    if(signal)
     g_signal_connect(toolitem, "clicked", GTK_SIGNAL_FUNC(signal), userdata);
    gtk_widget_set_sensitive(GTK_WIDGET(toolitem), bSensitive && signal);
    if(sTooltip) gtk_tool_item_set_tooltip(toolitem, tooltips, sTooltip, NULL);
    return toolitem;
}

GtkWidget* check_menu_item_new(const gchar* sText
 , const gchar* sPath, const gchar* sKey, GdkModifierType accel_mods
 , GCallback signal, gboolean bSensitive, gboolean bActive, CBrowser* browser)
{
    DEBUG_FUNC
    GtkWidget* menuitem = gtk_check_menu_item_new_with_mnemonic(sText);
    gtk_widget_set_sensitive(menuitem, bSensitive && signal != NULL);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menuitem), bActive);
    if(signal)
     g_signal_connect(menuitem, "activate", G_CALLBACK(signal), browser);
    sokoke_menu_item_set_accel(GTK_MENU_ITEM(menuitem), sPath, sKey, accel_mods);
    return menuitem;
}

GtkWidget* radio_button_new(GtkRadioButton* radio_button, const gchar* sLabel)
{
    return gtk_radio_button_new_with_mnemonic_from_widget(radio_button, sLabel);
}

void entry_setup_completion(GtkEntry* entry)
{
    DEBUG_FUNC
    /* TODO: The current behavior works only with the beginning of strings
             But we want to match "localhost" with "loc" and "hos" */
    GtkEntryCompletion* completion = gtk_entry_completion_new();
    gtk_entry_completion_set_model(completion
     , GTK_TREE_MODEL(gtk_list_store_new(1, G_TYPE_STRING)));
    gtk_entry_completion_set_text_column(completion, 0);
    gtk_entry_completion_set_minimum_key_length(completion, 3);
    gtk_entry_set_completion(entry, completion);
}

void entry_completion_append(GtkEntry* entry, const gchar* sText)
{
    DEBUG_FUNC
    GtkEntryCompletion* completion = gtk_entry_get_completion(entry);
    GtkTreeModel* completion_store = gtk_entry_completion_get_model(completion);
    GtkTreeIter iter;
    gtk_list_store_insert(GTK_LIST_STORE(completion_store), &iter, 0);
    gtk_list_store_set(GTK_LIST_STORE(completion_store), &iter, 0, sText, -1);
}

GtkWidget* get_current_document(CBrowser* browser)
{
    gint page = gtk_notebook_get_current_page(GTK_NOTEBOOK(browser->documents));
    GtkWidget* scrolled = gtk_notebook_get_nth_page(GTK_NOTEBOOK(browser->documents), page);
    return gtk_bin_get_child(GTK_BIN(scrolled));
}

CBrowser* document_get_browser(GtkWidget* document)
{
    // Find the browser for a document
    CBrowser* browser;
    GList* browsers = g_list_first(browser_list);
    do
    {
     browser = (CBrowser*)browsers->data;
     if(browser->document == document)
      return browser;
    }
    while(browsers = g_list_next(browsers));
    return NULL;
}

gint get_document_index(GtkWidget* document, CBrowser* browser)
{
    GtkWidget* scrolled = gtk_widget_get_parent(document);
    return gtk_notebook_page_num(GTK_NOTEBOOK(browser->documents), scrolled);
}

void update_favicon(CBrowser* browser)
{
    DEBUG_FUNC
    if(0) //browser->favicon // Has favicon?
    {
     // TODO: use custom icon
     // gtk_image_set_from_file(GTK_IMAGE(browser->icon_page), "image");
    }
    else if(0) // Known mime-type?
    {
     // TODO: Retrieve mime type and load icon; don't forget ftp listings
    }
    else
     gtk_image_set_from_stock(GTK_IMAGE(browser->document_icon), GTK_STOCK_FILE
      , GTK_ICON_SIZE_MENU);
}

void update_security(CBrowser* browser)
{
    DEBUG_FUNC
    // TODO: these are mozilla values, change that!
    //       0 = unknown, 4 = default, 262146 = secure
    // TODO: highlighted entry indicates security, find an alternative
    if(browser->iSecurity == 262146)
    {
     gtk_widget_modify_base(browser->navibar_location, GTK_STATE_NORMAL
      , &browser->navibar_location->style->base[GTK_STATE_SELECTED]);
     gtk_widget_modify_text(browser->navibar_location, GTK_STATE_NORMAL
      , &browser->navibar_location->style->text[GTK_STATE_SELECTED]);
     gtk_image_set_from_stock(GTK_IMAGE(browser->icon_lock)
      , GTK_STOCK_DIALOG_AUTHENTICATION, GTK_ICON_SIZE_MENU);
    }
    else
    {
     gtk_widget_modify_base(browser->navibar_location, GTK_STATE_NORMAL
      , NULL);
     gtk_widget_modify_text(browser->navibar_location, GTK_STATE_NORMAL
      , NULL);
     gtk_image_set_from_stock(GTK_IMAGE(browser->icon_lock)
      , GTK_STOCK_INFO, GTK_ICON_SIZE_MENU);
    }
}

void update_visibility(CBrowser* browser, gboolean visibility)
{
    DEBUG_FUNC
    if(gtk_notebook_get_n_pages(GTK_NOTEBOOK(browser->documents)) > 1
     || 1) return;
    // TODO: Check if this is a scripted window, else just return

    // SHOULD SCRIPTS BE ABLE TO HIDE WINDOWS AT ALL?
    if(0 && !visibility)
    {
     gtk_widget_hide(browser->window);
     return;
    }
    else if(!visibility)
     g_print("Window was not hidden.\n");

    sokoke_widget_set_visible(browser->menubar, browser->bMenubar);
    sokoke_widget_set_visible(browser->navibar, browser->bToolbar);
    sokoke_widget_set_visible(browser->navibar_location, browser->bLocation);
    sokoke_widget_set_visible(browser->navibar_search, browser->bLocation);
    sokoke_widget_set_visible(browser->statusbar, browser->bStatusbar);
    gtk_window_set_resizable(GTK_WINDOW(browser->window), browser->bResizable);
}

void action_set_active(const gchar* sAction, gboolean bActive
 , CBrowser* browser)
{
    DEBUG_FUNC
    // This shortcut toggles activity state by an action name
    GtkAction* action = gtk_action_group_get_action(browser->actiongroup, sAction);
    if(!GTK_IS_ACTION(action)){g_print("no action:<%s>\n", sAction);return;}//...
    gtk_toggle_action_set_active(GTK_TOGGLE_ACTION(action), bActive);
}

void action_set_sensitive(const gchar* sAction, gboolean bSensitive
 , CBrowser* browser)
{
    DEBUG_FUNC
    // This shortcut toggles sensitivity by an action name
    GtkAction* action = gtk_action_group_get_action(browser->actiongroup, sAction);
    if(!GTK_IS_ACTION(action)){g_print("no action:<%s>\n", sAction);return;}//...
    gtk_action_set_sensitive(action, bSensitive);
}

void action_set_visible(const gchar* sAction, gboolean bVisible
 , CBrowser* browser)
{
    DEBUG_FUNC
    // This shortcut toggles visibility by an action name
    GtkAction* action = gtk_action_group_get_action(browser->actiongroup, sAction);
    if(!GTK_IS_ACTION(action)){g_print("no action:<%s>\n", sAction);return;}//...
    gtk_action_set_visible(action, bVisible);
}

void update_statusbar_text(CBrowser* browser)
{
    DEBUG_FUNC
    if(browser->sStatusMessage)
    {
     gtk_statusbar_pop(GTK_STATUSBAR(browser->statusbar), 1);
     gtk_statusbar_push(GTK_STATUSBAR(browser->statusbar), 1
      , browser->sStatusMessage);
    }
    else
    {
     gchar message[56];
     if(browser->iLoadedPercent)
      g_snprintf(message, 55, "d%% loaded, %d/%d bytes"
       , browser->iLoadedPercent, browser->iLoadedBytes, browser->iLoadedBytesMax);
     else if(browser->iLoadedBytes)
      g_snprintf(message, 55, "%d bytes", browser->iLoadedBytes);
     else
      g_sprintf(message, " ");
     gtk_progress_bar_set_text(GTK_PROGRESS_BAR(browser->progress), message);
    }
}

void update_gui_state(CBrowser* browser)
{
    DEBUG_FUNC
    GtkWidget* document = get_current_document(browser);
    action_set_sensitive("ZoomIn", document_can_increase_text_size(document), browser);
    action_set_sensitive("ZoomOut", document_can_decrease_text_size(document), browser);
    action_set_sensitive("ZoomNormal", document_get_text_size(document) != 1, browser);
    action_set_sensitive("Back", document_can_go_back(document), browser);
    action_set_sensitive("Forward", document_can_go_forward(document), browser);
    action_set_sensitive("Refresh", browser->iLoadedPercent == -1, browser);
    action_set_sensitive("Stop", browser->iLoadedPercent != -1, browser);

    GtkAction* action = gtk_action_group_get_action(browser->actiongroup, "RefreshStop");
    if(browser->iLoadedPercent == -1)
    {
     g_object_set(action, "stock-id", GTK_STOCK_REFRESH, NULL);
     g_object_set(action, "tooltip", "Refresh the current page", NULL);
    }
    else
    {
     g_object_set(action, "stock-id", GTK_STOCK_STOP, NULL);
     g_object_set(action, "tooltip", "Stop loading the current page", NULL);
    }

    gtk_image_set_from_stock(GTK_IMAGE(browser->icon_page), GTK_STOCK_FILE
     , GTK_ICON_SIZE_MENU);

    if(browser->iLoadedBytesMax < 1) // Skip null and negative values
    {
     gtk_progress_bar_pulse(GTK_PROGRESS_BAR(browser->progress));
     update_statusbar_text(browser);
    }
    else
    {
     if(browser->iLoadedBytes > browser->iLoadedBytesMax)
      browser->iLoadedPercent = 100;
     else
      browser->iLoadedPercent = (browser->iLoadedBytes * 100) / browser->iLoadedBytesMax;
     gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(browser->progress)
      , browser->iLoadedPercent / 100.0);
     update_statusbar_text(browser);
    }
}

void update_feeds(CBrowser* browser)
{
    // TODO: Look for available feeds, requires dom access
}

void update_search_engines(CBrowser* browser)
{
    // TODO: Look for available search engines, requires dom access
}

void update_status_message(CBrowser* browser, const gchar* message)
{
    DEBUG_FUNC
    GString* string = g_string_new(message);
    browser->sStatusMessage = g_string_free(string, FALSE);
    update_statusbar_text(browser);
}

void update_browser_actions(CBrowser* browser)
{
    DEBUG_FUNC
    gboolean bActive = gtk_notebook_get_n_pages(
     GTK_NOTEBOOK(browser->documents)) > 1 ? TRUE : FALSE;
    gtk_notebook_set_show_tabs(GTK_NOTEBOOK(browser->documents), bActive);
    action_set_sensitive("TabClose", bActive, browser);
    GList* children = gtk_container_get_children(
     GTK_CONTAINER(browser->menu_tabsclosed));
    action_set_sensitive("UndoTabClose", g_list_length(children) > 3, browser);
    action_set_sensitive("TabsClosed", g_list_length(children) > 3, browser);
    g_list_free(children);
}

gchar* magic_uri(const gchar* sUri)
{
    // Add file:// if we have a local path
    if(g_path_is_absolute(sUri)) return g_strconcat("file://", sUri, NULL);
    // Do we need to add a protocol?
    if(!strstr(sUri, "://"))
    {
     // Do we have a domain, ip address or localhost?
     if(strstr(sUri, ".") != NULL || !strcmp(sUri, "localhost"))
      return g_strconcat("http://", sUri, NULL);
     // Do we have a keyword and a string?
     if(strstr(sUri, " ") != NULL)
      ; // TODO: Search, using smart keywords
     // We only have a word, so search for it
     gchar sSearch[MIDORI_LEN_URI + 1];
     g_snprintf(sSearch, MIDORI_LEN_URI, config->sLocationSearch, sUri);
     return g_strdup(sSearch);
    }
    return g_strdup(sUri);
}

gchar* get_default_font(void)
{
    GtkSettings* gtksettings = gtk_settings_get_default();
    gchar* sDefaultFont;
    g_object_get(gtksettings, "gtk-font-name", &sDefaultFont, NULL);
    return sDefaultFont;
}

GtkToolbarStyle config_to_toolbarstyle(guint iToolbarStyle)
{
    switch(iToolbarStyle)
    {
     case 1:
      return GTK_TOOLBAR_ICONS;
     case 2:
      return GTK_TOOLBAR_TEXT;
     case 3:
      return GTK_TOOLBAR_BOTH;
     case 4:
      return GTK_TOOLBAR_BOTH_HORIZ;
    }
    GtkSettings* gtksettings = gtk_settings_get_default();
    g_object_get(gtksettings, "gtk-toolbar-style", &iToolbarStyle, NULL);
    return iToolbarStyle;
}

GtkToolbarStyle config_to_toolbariconsize(guint iToolbarIconSize)
{
    return iToolbarIconSize
     ? GTK_ICON_SIZE_SMALL_TOOLBAR
     : GTK_ICON_SIZE_LARGE_TOOLBAR;
}

// -- Common (pseudo) signal handlers begin here

void on_document_location_changed(GtkWidget* document
#ifdef MIDORI_WEBKIT
 , const gchar* sUri
#endif
, CBrowser* browser)
{
    DEBUG_FUNC
    #ifndef MIDORI_WEBKIT
    const gchar* sUri = document_get_uri(document);
    #endif
    browser->sUri = g_strdup(sUri);
    if(document == get_current_document(browser))
    {
     if(browser->sUri)
     {
      gtk_entry_set_text(GTK_ENTRY(browser->navibar_location), browser->sUri);
      gtk_label_set_text(GTK_LABEL(browser->document_name), browser->sUri);
     }
     update_status_message(browser, NULL);
     update_gui_state(browser);
    }
}

void on_document_title_changed(GtkWidget* document
#ifdef MIDORI_WEBKIT
 , const gchar* sTitle, const gchar* sUri
#endif
 , CBrowser* browser)
{
    #ifdef MIDORI_WEBKIT
    // TODO: We emulate location_changed here for now
    // Shouldn't we have separated title_changed and location_changed signals?
    on_document_location_changed(document, sUri, browser);
    #endif
    DEBUG_FUNC
    #ifndef MIDORI_WEBKIT
    const gchar* sTitle = document_get_title(document);
    #endif
    browser->sTitle = g_strdup(sTitle);
    if(browser->sTitle)
    {
     gtk_label_set_text(GTK_LABEL(browser->document_name), browser->sTitle);
     // TODO: Are tab label tooltips always good?
     gtk_tooltips_set_tip(GTK_TOOLTIPS(tooltips), gtk_widget_get_parent(
      gtk_widget_get_parent(browser->document_name)), browser->sTitle, NULL);
     gtk_label_set_text(GTK_LABEL(gtk_bin_get_child(GTK_BIN(
      browser->document_menu))), browser->sTitle);
     if(document == get_current_document(browser))
     {
      gchar* sWindowTitle = g_strconcat(browser->sTitle, " - ", PACKAGE_NAME, NULL);
      gtk_window_set_title(GTK_WINDOW(browser->window), sWindowTitle);
      g_free(sWindowTitle);
     }
    }
}

#ifdef MIDORI_WEBKIT
void on_frame_title_changed(WebKitGtkFrame* frame
 , const gchar* sTitle, const gchar* sUri, CBrowser* browser)
{
    DEBUG_FUNC
    UNIMPLEMENTED
}

void on_document_icon_changed(GtkWidget* document, WebKitGtkFrame* widget
 , CBrowser* browser)
{
    DEBUG_FUNC
    // TODO: Implement icon updates; currently this isn't ever called anyway
    const gchar* sIcon = NULL;
    UNIMPLEMENTED
    if(sIcon)
    {
     gtk_label_set_text(GTK_LABEL(browser->document_name), "icon");
     gtk_label_set_text(GTK_LABEL(gtk_bin_get_child(GTK_BIN(
      browser->document_menu))), "icon");
     if(document == get_current_document(browser))
     {
      gchar* sWindowTitle = g_strconcat("icon", " - ", PACKAGE_NAME, NULL);
      gtk_window_set_title(GTK_WINDOW(browser->window), sWindowTitle);
      g_free(sWindowTitle);
     }
    }
}
#endif

void on_document_load_started(GtkWidget* document
#ifdef MIDORI_WEBKIT
 , WebKitGtkFrame* widget
#endif
, CBrowser* browser)
{
    DEBUG_FUNC
    browser->iLoadedPercent = 0;
    browser->iLoadedBytes = 0;
    browser->iLoadedBytesMax = 0;
    update_favicon(browser);
    if(document == get_current_document(browser)) update_gui_state(browser);
    update_statusbar_text(browser);
    gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(browser->progress), 0.1);
    gtk_widget_show(browser->progress);
}

void on_document_load_changed(GtkWidget* document
#ifdef MIDORI_WEBKIT
 , gint progress
#else
 , const WebiLoadStatus* status
#endif
 , CBrowser* browser)
{
    DEBUG_FUNC
    #ifndef MIDORI_WEBKIT
    browser->iLoadedBytes = status->received;
   browser->iLoadedBytesMax = status->size;
    #else
    browser->iLoadedBytes = progress;
    browser->iLoadedBytes = 100;
    #endif
    if(document == get_current_document(browser)) update_gui_state(browser);
}

void on_document_load_finished(GtkWidget* document
#ifdef MIDORI_WEBKIT
 , WebKitGtkFrame* widget
#endif
, CBrowser* browser)
{
    DEBUG_FUNC
    browser->iLoadedPercent = -1;
    update_favicon(browser);
    if(document == get_current_document(browser)) update_gui_state(browser);
    update_statusbar_text(browser);
    gtk_widget_hide(browser->progress);
}

#ifdef MIDORI_WEBKIT
void on_frame_cleared(WebKitGtkFrame* frame, CBrowser* browser)
{
    DEBUG_FUNC
    UNIMPLEMENTED
}

void on_frame_load_done(WebKitGtkFrame* frame, gboolean bSuccess
 , CBrowser* browser)
{
    DEBUG_FUNC
    UNIMPLEMENTED
}
#endif

void on_document_status_message(GtkWidget* document, gchar* sText
 , CBrowser* browser)
{
    DEBUG_FUNC
    update_status_message(browser, sText);
}

void on_document_selection_changed(GtkWidget* document, CBrowser* browser)
{
    //g_print("%s\n", "selection_changed");
}

void on_document_request_script_prompt(GtkWidget* document
#ifdef MIDORI_WEBI
 , WebiPromptArgs* args
#endif
 , CBrowser* browser)
{
    DEBUG_FUNC
    GtkDialog* dialog;
    gchar* title;
    GtkMessageType type = GTK_MESSAGE_OTHER;
    GtkButtonsType buttons = GTK_BUTTONS_NONE;
    gint default_response = GTK_RESPONSE_NONE;
    gint response;
    GtkWidget* widget = NULL;

    #ifdef MIDORI_WEBI
    switch(args->type)
    {
     case WEBI_ALERT:
      type = GTK_MESSAGE_WARNING;
      buttons = GTK_BUTTONS_CLOSE;
      default_response = GTK_RESPONSE_CLOSE; break;
     case WEBI_CONFIRM:
      type = GTK_MESSAGE_QUESTION;
      buttons = GTK_BUTTONS_YES_NO;
      default_response = GTK_RESPONSE_YES; break;
     case WEBI_INPUT:
      type = GTK_MESSAGE_QUESTION;
      buttons = GTK_BUTTONS_OK_CANCEL;
      default_response = GTK_RESPONSE_OK;
    }
    #endif

    dialog = GTK_DIALOG(gtk_message_dialog_new(GTK_WINDOW(browser->window)
     , GTK_DIALOG_DESTROY_WITH_PARENT, type, buttons
    #ifdef MIDORI_WEBI
     , args->msg
    #else
     , ""
    #endif
    ));
    title = g_strconcat("JavaScript - ", browser->sUri, NULL);
    gtk_window_set_title(GTK_WINDOW(dialog), title); g_free(title);

    #ifdef MIDORI_WEBI
    if(args->type == WEBI_INPUT)
    #else
    if(0)
    #endif
    {
     widget = gtk_entry_new();
     #ifdef MIDORI_WEBI
     gtk_entry_set_text(GTK_ENTRY(widget), args->default_input);
     #endif
     gtk_container_add(GTK_CONTAINER(dialog->vbox), widget);
     gtk_widget_show(widget);
    }

    gtk_dialog_set_default_response(dialog, default_response);
    gtk_widget_set_sensitive(document, FALSE);
    response = sokoke_dialog_run_modeless(dialog);
    gtk_widget_set_sensitive(document, TRUE);

    #ifdef MIDORI_WEBI
    switch(response)
    {
     case GTK_RESPONSE_YES:
      args->out_ok_pressed = TRUE; break;
     case GTK_RESPONSE_NO:
      args->out_ok_pressed = FALSE; break;
     case GTK_RESPONSE_OK:
      args->out_input = g_strdup(gtk_entry_get_text(GTK_ENTRY(widget)));
    }
    #endif
    gtk_widget_destroy(GTK_WIDGET(dialog));
}

void on_document_request_auth_prompt(GtkWidget* document
#ifdef MIDORI_WEBI
 , WebiAuthArgs* args
#endif
 , CBrowser* browser)
{
    DEBUG_FUNC
    GtkDialog* dialog;
    gchar* title; gchar* msg;
    GtkWidget* hbox;
    GtkSizeGroup* sizegroup; GtkWidget* align; GtkWidget* label;
    GtkWidget* user; GtkWidget* pass;
    gint response;

    #ifdef MIDORI_WEBI
    msg = g_strdup_printf("Authentication required for:\n\n%s", args->realm);
    #endif
    dialog = GTK_DIALOG(gtk_message_dialog_new(GTK_WINDOW(browser->window)
     , GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_OTHER, GTK_BUTTONS_OK_CANCEL
     , msg)); g_free(msg);
    #ifdef MIDORI_WEBI
    title = g_strconcat("Authentication - ", args->url, NULL);
    #endif
    gtk_window_set_title(GTK_WINDOW(dialog), title); g_free(title);
    #if GTK_CHECK_VERSION(2, 10, 0)
    GtkWidget* image = gtk_image_new_from_icon_name(
     GTK_STOCK_DIALOG_AUTHENTICATION, GTK_ICON_SIZE_DIALOG);
    gtk_widget_show(image);
    gtk_message_dialog_set_image(GTK_MESSAGE_DIALOG(dialog), image);
    #endif
    sizegroup = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);
    hbox = gtk_hbox_new(FALSE, 6);
    label = gtk_label_new("Username");
    align = gtk_alignment_new(0, 0.5, 0, 0);
    gtk_container_add(GTK_CONTAINER(align), label);
    gtk_size_group_add_widget(sizegroup, align);
    gtk_box_pack_start(GTK_BOX(hbox), align, FALSE, FALSE, 0);
    user = gtk_entry_new();
    gtk_box_pack_start(GTK_BOX(hbox), user, FALSE, FALSE, 0);
    gtk_container_add(GTK_CONTAINER(dialog->vbox), hbox);
    gtk_widget_show_all(hbox);
    hbox = gtk_hbox_new(FALSE, 6);
    label = gtk_label_new("Password");
    align = gtk_alignment_new(0, 0.5, 0, 0);
    gtk_container_add(GTK_CONTAINER(align), label);
    gtk_size_group_add_widget(sizegroup, align);
    gtk_box_pack_start(GTK_BOX(hbox), align, FALSE, FALSE, 0);
    pass = gtk_entry_new();
    gtk_entry_set_visibility(GTK_ENTRY(pass), FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), pass, FALSE, FALSE, 0);
    gtk_container_add(GTK_CONTAINER(dialog->vbox), hbox);
    gtk_widget_show_all(hbox);
    gtk_dialog_set_default_response(dialog, GTK_RESPONSE_CANCEL);

    gtk_widget_set_sensitive(document, FALSE);
    response = sokoke_dialog_run_modeless(dialog);
    gtk_widget_set_sensitive(document, TRUE);

    #ifdef MIDORI_WEBI
    if(response == GTK_RESPONSE_OK)
    {
     args->out_username = g_strdup(gtk_entry_get_text(GTK_ENTRY(user)));
     args->out_password = g_strdup(gtk_entry_get_text(GTK_ENTRY(pass)));
     args->out_ok_pressed = TRUE;
    }
    else
     args->out_ok_pressed = FALSE;
    #endif
    gtk_widget_destroy(GTK_WIDGET(dialog));
}

gboolean on_document_set_cookie(GtkWidget* document
#ifdef MIDORI_WEBI
 , WebiCookie* cookie
#endif
 , CBrowser* browser)
{
    DEBUG_FUNC
    // Whole cookie
    // const gchar* cookie;

    // Cookie name
    // const gchar* name;

    // Cookie value
    // const gchar* value;

    // Cookie comment
    // const gchar* comment;

    // Cookie domain
    // const gchar* domain;

    // Cookie TTL
    // const gchar* maxAge;

    // Cookie path
    // const gchar* path;

    // Is cookie secure?
    // gboolean secure;

    // Cookie-specification version.
    // gint version;

    return TRUE;
}

void on_document_link_hover(GtkWidget* document, const gchar* sTitle
 , const gchar* sLabel, const gchar* sUri, const gchar* sTarget, CBrowser* browser)
{
    DEBUG_FUNC
    update_status_message(browser, sUri);
    sokoke_widget_set_visible(browser->icon_link_window
     , sTarget && !g_ascii_strncasecmp(sTarget, "_blank", 6));
    // TODO: detect "onclick", "onmousedown", "onmouseup" javascript links
    sokoke_widget_set_visible(browser->icon_link_script
     , sUri && g_str_has_prefix(sUri, "javascript:"));
}

GtkWidget* on_document_window_open(GtkWidget* document, const gchar* sUri
 , CBrowser* browser)
{
    DEBUG_FUNC
    // A window is created
    // TODO: Find out if this comes from a script or a click
    // TODO: Block scripted popups, return NULL and show status icon
    // Shoud we open a new window or a new tab?
    gboolean bWindow;
    switch(config->iPopupRule)
    {
     case 0: // All popups
         bWindow = TRUE; break;
     case 1: // Only customized popups
         // TODO: Check if window has non-default features
         bWindow = TRUE; break;
     default: // case 2: // No new windows
         bWindow = FALSE;
    }
    CBrowser* newBrowser = browser_new(bWindow ? NULL : browser);
    /*webi_set_group(WEBI(newBrowser->document)
     , webi_get_group(get_current_document(browser)));*/
    return newBrowser->document;
}

void on_document_window_features(GtkWidget* document
#ifdef MIDORI_WEBI
 , WebiWindowProperties* properties
#endif
 , CBrowser* browser)
{
    DEBUG_FUNC
    #ifdef MIDORI_WEBI
    browser->bMenubar = TRUE;
    browser->bToolbar = properties->toolbarsVisible;
    browser->bLocation = properties->toolbarsVisible;
    browser->bStatusbar = properties->statusBarVisible;
    //browser->bScrollbars = properties->scrollbarsVisible;
    browser->bResizable = properties->isResizable;
    #endif
    update_visibility(browser, TRUE);
}

// Document actions begin here

#define DOCUMENT_IS_DEFINED(rv) \
 if(!GTK_IS_WIDGET(document)) return rv
#define DOCUMENT_IS_SENSITIVE(rv) \
 if(!GTK_WIDGET_SENSITIVE(document)) return rv

void document_load_uri(GtkWidget* document, const gchar* sUri)
{
    DEBUG_FUNC
    DOCUMENT_IS_SENSITIVE();
    // TODO: Check if we want an external app for this uri scheme
    // TODO: This isn't the place for uri scheme handling, is it?
    #ifdef MIDORI_WEBI
    webi_load_url(WEBI(document), sUri);
    #elif defined(MIDORI_WEBKIT)
    // TODO: The block below is just for testing :D
    if(sUri && g_str_has_prefix(sUri, "res://"))
    {
     webkit_gtk_page_load_html_string(WEBKIT_GTK_PAGE(document)
      , "<html><head><title>" PACKAGE_STRING "</title></head>"
        "<body>test</body></html>"
      , "about:about");
     return;
    }
    webkit_gtk_page_open(WEBKIT_GTK_PAGE(document), (gchar*)sUri);
    #endif
}

void document_go_back(GtkWidget* document)
{
    DEBUG_FUNC
    DOCUMENT_IS_SENSITIVE();
    #ifdef MIDORI_WEBI
    webi_go_back(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    webkit_gtk_page_go_backward(WEBKIT_GTK_PAGE(document));
    #endif
}

void document_go_forward(GtkWidget* document)
{
    DEBUG_FUNC
    DOCUMENT_IS_SENSITIVE();
    #ifdef MIDORI_WEBI
    webi_go_forward(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    webkit_gtk_page_go_forward(WEBKIT_GTK_PAGE(document));
    #endif
}

void document_refresh(GtkWidget* document, gboolean bSkipCache)
{
    DEBUG_FUNC
    DOCUMENT_IS_SENSITIVE();
    #ifdef MIDORI_WEBI
    webi_refresh(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    webkit_gtk_page_reload(WEBKIT_GTK_PAGE(document));
    #endif
}

void document_stop(GtkWidget* document)
{
    DEBUG_FUNC
    DOCUMENT_IS_SENSITIVE();
    #ifdef MIDORI_WEBI
    webi_stop_load(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    webkit_gtk_page_stop_loading(WEBKIT_GTK_PAGE(document));
    #endif
}

gboolean document_can_go_back(GtkWidget* document)
{
    DEBUG_FUNC
    DOCUMENT_IS_DEFINED(FALSE);
    #ifdef MIDORI_WEBI
    return webi_can_go_back(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    return webkit_gtk_page_can_go_backward(WEBKIT_GTK_PAGE(document));
    #endif
}

gboolean document_can_go_forward(GtkWidget* document)
{
    DEBUG_FUNC
    DOCUMENT_IS_DEFINED(FALSE);
    #ifdef MIDORI_WEBI
    return webi_can_go_forward(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    return webkit_gtk_page_can_go_forward(WEBKIT_GTK_PAGE(document));
    #endif
}

G_CONST_RETURN gchar* document_get_uri(GtkWidget* document)
{
    DEBUG_FUNC
    DOCUMENT_IS_DEFINED(NULL);
    #ifdef MIDORI_WEBI
    return webi_get_location(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED return ""; //webkit_gtk_page_get_uri(WEBKIT_GTK_PAGE(document));
    #endif
}

G_CONST_RETURN gchar* document_get_title(GtkWidget* document)
{
    DEBUG_FUNC
    DOCUMENT_IS_DEFINED(NULL);
    #ifdef MIDORI_WEBI
    return webi_get_title(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED return ""; //webkit_gtk_page_get_title(WEBKIT_GTK_PAGE(document));
    #endif
}

// FIXME: This better be document_find(document, text, forward, case, wrap)
gboolean document_find(GtkWidget* document, const gchar* sText
 , gboolean bCase, gboolean bForward)
{
    DEBUG_FUNC
    DOCUMENT_IS_SENSITIVE(FALSE);
    #ifdef MIDORI_WEBI
    return webi_find(WEBI(document), sText, bCase, bForward);
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED return FALSE; /*webkit_gtk_page_search_for(WEBKIT_GTK_PAGE(document)
     , sText, bForward, bCase, TRUE);*/
    #endif
}

gboolean document_can_increase_text_size(GtkWidget* document)
{
    DEBUG_FUNC
    DOCUMENT_IS_DEFINED(FALSE);
    DOCUMENT_IS_SENSITIVE(FALSE);
    #ifdef MIDORI_WEBI
    return TRUE;
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED return FALSE; //webkit_gtk_page_can_increase_text_size(WEBKIT_GTK_PAGE(document));
    #endif
}

gboolean document_can_decrease_text_size(GtkWidget* document)
{
    DEBUG_FUNC
    DOCUMENT_IS_DEFINED(FALSE);
    DOCUMENT_IS_SENSITIVE(FALSE);
    #ifdef MIDORI_WEBI
    return TRUE;
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED return FALSE; //webkit_gtk_page_can_decrease_text_size(WEBKIT_GTK_PAGE(document));
    #endif
}

void document_set_text_size(GtkWidget* document, gfloat iSize)
{
    DEBUG_FUNC
    DOCUMENT_IS_SENSITIVE();
    #ifdef MIDORI_WEBI
    webi_set_text_multiplier(WEBI(document), iSize);
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED //webkit_gtk_page_set_text_size_multiplier(WEBKIT_GTK_PAGE(document), iSize);
    #endif
}

gfloat document_get_text_size(GtkWidget* document)
{
    DEBUG_FUNC
    DOCUMENT_IS_DEFINED(1);
    #ifdef MIDORI_WEBI
    return webi_get_text_multiplier(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED return 1;//webkit_gtk_page_get_text_size_multiplier(WEBKIT_GTK_PAGE(document));
    #endif
}

void document_close(GtkWidget* document, CBrowser* browser)
{
    DEBUG_FUNC
    // TODO: Delete old items if > 10 items
    GtkWidget* menuitem = menu_item_new(browser->sTitle
     , GTK_STOCK_FILE, "", NULL, GDK_CONTROL_MASK
     , G_CALLBACK(on_menu_tabsclosed_item_activate), TRUE, browser);
    gtk_tooltips_set_tip(GTK_TOOLTIPS(tooltips), menuitem, browser->sUri, NULL);
    g_object_set_data(G_OBJECT(menuitem), "uri", browser->sUri);
    gtk_menu_shell_prepend(GTK_MENU_SHELL(browser->menu_tabsclosed), menuitem);
    gtk_widget_show(menuitem);
    gtk_widget_destroy(browser->document_menu);
    gtk_notebook_remove_page(GTK_NOTEBOOK(browser->documents)
     , get_document_index(document, browser));
    update_browser_actions(browser);
}

#undef DOCUMENT_IS_DEFINED
#undef DOCUMENT_IS_SENSITIVE

// -- GTK+ signal handlers begin here

void on_action_window_new_activate(GtkAction* action, CBrowser* browser)
{
    browser_new(NULL);
}

void on_action_tab_new_activate(GtkAction* action, CBrowser* browser)
{
    browser_new(browser);
    update_browser_actions(browser);
}

void on_action_tab_close_activate(GtkAction* action, CBrowser* browser)
{
    DEBUG_FUNC
    document_close(get_current_document(browser), browser);
}

void on_action_window_close_activate(GtkAction* action, CBrowser* browser)
{
    gtk_widget_destroy(GTK_WIDGET(browser->window));
}

void on_action_quit_activate(GtkAction* action, CBrowser* browser)
{
    gtk_main_quit();
}

void on_action_find_activate(GtkAction* action, CBrowser* browser)
{
    if(GTK_WIDGET_VISIBLE(browser->findbox))
     gtk_widget_hide(browser->findbox);
    else
    {
     gtk_widget_show(browser->findbox);
     gtk_widget_grab_focus(GTK_WIDGET(browser->findbox_text));
    }
}

void on_action_find_next_activate(GtkAction* action, CBrowser* browser)
{
    if(!GTK_WIDGET_VISIBLE(browser->findbox))
     ; // TODO: What if the findbox is hidden?
    const gchar* sText = gtk_entry_get_text(GTK_ENTRY(browser->findbox_text));
    const gboolean bCase = gtk_toggle_tool_button_get_active(
     GTK_TOGGLE_TOOL_BUTTON(browser->findbox_case));
    document_find(get_current_document(browser), sText, bCase, TRUE);
}

void on_action_find_previous_activate(GtkAction* action, CBrowser* browser)
{
    if(!GTK_WIDGET_VISIBLE(browser->findbox))
     ; // TODO: What if the findbox is hidden?
    const gchar* sText = gtk_entry_get_text(GTK_ENTRY(browser->findbox_text));
    const gboolean bCase = gtk_toggle_tool_button_get_active(
     GTK_TOGGLE_TOOL_BUTTON(browser->findbox_case));
    document_find(get_current_document(browser), sText, bCase, FALSE);
}

gboolean on_settings_homepage_focus_out(GtkWidget* widget, GdkEventFocus event
 , CBrowser* browser)
{
    g_free(config->sHomepage);
    config->sHomepage = g_strdup(gtk_entry_get_text(GTK_ENTRY(widget)));
    return FALSE;
}

void on_settings_toolbarstyle_changed(GtkWidget* widget, CBrowser* browser)
{
    config->iToolbarStyle = gtk_combo_box_get_active(GTK_COMBO_BOX(widget));
    gtk_toolbar_set_style(GTK_TOOLBAR(browser->navibar)
     , config_to_toolbarstyle(config->iToolbarStyle));
}

void on_settings_toolbariconsize_toggled(GtkWidget* widget, CBrowser* browser)
{
    config->iToolbarIconSize = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(browser->navibar)
     , config_to_toolbariconsize(config->iToolbarIconSize));
}

void on_settings_closebuttonsontabs_toggled(GtkWidget* widget, CBrowser* browser)
{
    config->iTabClose = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
    GList* browsers = browser_list;
    CBrowser* curBrowser;
    do
    {
     curBrowser = (CBrowser*)browsers->data;
     sokoke_widget_set_visible(curBrowser->document_close, config->iTabClose);
    }
    while(browsers = g_list_next(browsers)); g_list_free(browsers);
}

gboolean on_settings_locationsearch_focus_out(GtkWidget* widget, GdkEventFocus event
 , CBrowser* browser)
{
    g_free(config->sLocationSearch);
    config->sLocationSearch = g_strdup(gtk_entry_get_text(GTK_ENTRY(widget)));
    return FALSE;
}

void on_action_preferences_activate(GtkAction* action, CBrowser* browser)
{
    DEBUG_FUNC
    // Show the preferences dialog. Create it if necessary.
    if(GTK_IS_DIALOG(settings_dialog))
     gtk_window_present(GTK_WINDOW(settings_dialog));
    else
    {
     gchar* sDialogTitle = g_strdup_printf("%s Preferences", PACKAGE_NAME);
     settings_dialog = gtk_dialog_new_with_buttons(sDialogTitle
                                 , GTK_WINDOW(browser->window)
                                 , GTK_DIALOG_DESTROY_WITH_PARENT
                                 , GTK_STOCK_HELP
                                 , GTK_RESPONSE_HELP
                                 , GTK_STOCK_CLOSE
                                 , GTK_RESPONSE_CLOSE
                                 , NULL);
     gtk_window_set_icon_name(GTK_WINDOW(settings_dialog), GTK_STOCK_PREFERENCES);
     gtk_dialog_set_has_separator(GTK_DIALOG(settings_dialog), FALSE);
     // TODO: Implement some kind of help function
     gtk_dialog_set_response_sensitive(GTK_DIALOG(settings_dialog)
      , GTK_RESPONSE_HELP, FALSE); //...
     g_signal_connect(GTK_OBJECT(settings_dialog), "response"
      , GTK_SIGNAL_FUNC(gtk_widget_destroy), settings_dialog); //...
     // TODO: Do we want tooltips for explainations or can we omit that?
     // TODO: Connect widgets to some kind of instant-apply handler
     // TODO: We need mnemonics
     if(!strcmp(g_getenv("DESKTOP_SESSION"), "xfce"))
     {
      // Use an Xfce style header if running in XFCE
      GtkWidget* xfce_heading = gtk_event_box_new();
      gtk_widget_modify_bg(xfce_heading, GTK_STATE_NORMAL
       , &browser->navibar_location->style->base[GTK_STATE_NORMAL]);
      GtkWidget* hbox = gtk_hbox_new(FALSE, 12);
      gtk_container_set_border_width(GTK_CONTAINER(hbox), 6);
      GtkWidget* icon = gtk_image_new_from_icon_name(
       gtk_window_get_icon_name(GTK_WINDOW(browser->window))
       , GTK_ICON_SIZE_DIALOG);
      gtk_box_pack_start(GTK_BOX(hbox), icon, FALSE, FALSE, 0);
      GtkWidget* label = gtk_label_new(NULL);
      gtk_widget_modify_fg(label, GTK_STATE_NORMAL
       , &browser->navibar_location->style->text[GTK_STATE_NORMAL]);
      gchar* sTitle = g_strdup_printf("<span size='large' weight='bold'>%s</span>"
       , sDialogTitle);
      gtk_label_set_markup(GTK_LABEL(label), sTitle); g_free(sTitle);
      gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
      gtk_container_add(GTK_CONTAINER(xfce_heading), hbox);
      gtk_box_pack_start(GTK_BOX(GTK_DIALOG(settings_dialog)->vbox)
      , xfce_heading, FALSE, FALSE, 0);
     } g_free(sDialogTitle);
     GtkWidget* notebook = gtk_notebook_new();
     gtk_container_set_border_width(GTK_CONTAINER(notebook), 6);
     GtkSizeGroup* sizegroup = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);
     GtkSizeGroup* sizegroup2 = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);
     GtkWidget* page; GtkWidget* frame; GtkWidget* table; GtkWidget* align;
     GtkWidget* button; GtkWidget* checkbutton; GtkWidget* colorbutton;
     GtkWidget* combobox; GtkWidget* entry; GtkWidget* hbox; GtkWidget* spinbutton;
     #define PAGE_NEW(__label) page = gtk_vbox_new(FALSE, 0);\
      gtk_container_set_border_width(GTK_CONTAINER(page), 5);\
      gtk_notebook_append_page(GTK_NOTEBOOK(notebook), page, gtk_label_new(__label))
     #define FRAME_NEW(__label) frame = sokoke_hig_frame_new(__label);\
      gtk_container_set_border_width(GTK_CONTAINER(frame), 5);\
      gtk_box_pack_start(GTK_BOX(page), frame, FALSE, FALSE, 0);
     #define TABLE_NEW(__rows, __cols) table = gtk_table_new(__rows, __cols, FALSE);\
      gtk_container_set_border_width(GTK_CONTAINER(table), 5);\
      gtk_container_add(GTK_CONTAINER(frame), table);
     #define WIDGET_ADD(__widget, __left, __right, __top, __bottom)\
      gtk_table_attach(GTK_TABLE(table), __widget\
       , __left, __right, __top, __bottom\
       , 0, GTK_FILL, 8, 2)
     #define FILLED_ADD(__widget, __left, __right, __top, __bottom)\
      gtk_table_attach(GTK_TABLE(table), __widget\
       , __left, __right, __top, __bottom\
       , GTK_EXPAND | GTK_FILL, GTK_FILL, 8, 2)
     #define INDENTED_ADD(__widget, __left, __right, __top, __bottom)\
      align = gtk_alignment_new(0, 0.5, 0, 0);\
      gtk_container_add(GTK_CONTAINER(align), __widget);\
      gtk_size_group_add_widget(sizegroup, align);\
      WIDGET_ADD(align, __left, __right, __top, __bottom)
     #define SEMI_INDENTED_ADD(__widget, __left, __right, __top, __bottom)\
      align = gtk_alignment_new(0, 0.5, 0, 0);\
      gtk_container_add(GTK_CONTAINER(align), __widget);\
      gtk_size_group_add_widget(sizegroup2, align);\
      WIDGET_ADD(align, __left, __right, __top, __bottom)
     #define SPANNED_ADD(__widget, __left, __right, __top, __bottom)\
      align = gtk_alignment_new(0, 0.5, 0, 0);\
      gtk_container_add(GTK_CONTAINER(align), __widget);\
      FILLED_ADD(align, __left, __right, __top, __bottom)
     // Page "General"
     PAGE_NEW("General");
     FRAME_NEW("Startup");
     TABLE_NEW(2, 2);
     INDENTED_ADD(gtk_label_new("Load on startup"), 0, 1, 0, 1);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "Blank page", "Homepage", "Last session", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 1); //...
     gtk_widget_set_sensitive(combobox, FALSE); //...
     FILLED_ADD(combobox, 1, 2, 0, 1);
     INDENTED_ADD(gtk_label_new("Homepage"), 0, 1, 1, 2);
     entry = gtk_entry_new();
     gtk_entry_set_text(GTK_ENTRY(entry), config->sHomepage);
     g_signal_connect(entry, "focus-out-event"
     , G_CALLBACK(on_settings_homepage_focus_out), browser);
     FILLED_ADD(entry, 1, 2, 1, 2);
     // TODO: We need something like "use current website"
     FRAME_NEW("Downloads");
     TABLE_NEW(1, 2);
     INDENTED_ADD(gtk_label_new("Download folder"), 0, 1, 0, 1);
     GtkWidget* filebutton = gtk_file_chooser_button_new(
      "Choose download folder", GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER);
     gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(filebutton)
      , g_get_home_dir()); //...
     gtk_widget_set_sensitive(filebutton, FALSE); //...
     FILLED_ADD(filebutton, 1, 2, 0, 1);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Show a notification window for finished downloads");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 2, 1, 2);
     FRAME_NEW("Languages");
     TABLE_NEW(1, 2);
     INDENTED_ADD(gtk_label_new("Preferred languages"), 0, 1, 0, 1);
     entry = gtk_entry_new();
     // TODO: Make sth like get_browser_languages_default filtering encodings and C out
     // TODO: Provide a real ui with real language names (iso-codes)
     const gchar* const* sLanguages = g_get_language_names();
     //for (i = 0; sLanguages[i] != NULL; i++) g_strdup(sLanguages[i]);
     //for (str = strings; *str; str++);
     gchar* sLanguagesPreferred = g_strjoinv(",", (gchar**)sLanguages);
     gtk_entry_set_text(GTK_ENTRY(entry), sLanguagesPreferred/*config->sLanguagesPreferred*/);
     g_free(sLanguagesPreferred);
     gtk_widget_set_sensitive(entry, FALSE); //...
     FILLED_ADD(entry, 1, 2, 0, 1);
     // Page "Appearance"
     PAGE_NEW("Appearance");
     FRAME_NEW("Font settings");
     TABLE_NEW(5, 2);
     INDENTED_ADD(gtk_label_new("Standard font"), 0, 1, 0, 1);
     button = gtk_font_button_new_with_font("Sans 10"/*config->sFontStandard*/);
     gtk_widget_set_sensitive(button, FALSE); //...
     FILLED_ADD(button, 1, 2, 0, 1);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Use theme _default font");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 2, 1, 2);
     INDENTED_ADD(gtk_label_new("Minimum font size"), 0, 1, 2, 3);
     spinbutton = gtk_spin_button_new_with_range(6, 30, 1);
     gtk_spin_button_set_value(GTK_SPIN_BUTTON(spinbutton), 8/*config->iFontSizeMin*/);
     gtk_widget_set_sensitive(spinbutton, FALSE); //...
     SPANNED_ADD(spinbutton, 1, 2, 2, 3);
     INDENTED_ADD(gtk_label_new("Default encoding"), 0, 1, 3, 4);
     combobox = gtk_combo_box_new_text();
     const gchar* encoding = NULL; g_get_charset(&encoding);
     // TODO: Fallback to utf-8 if the encoding is not sane (e.g. when lang=C)
     gchar* sEncodingDefault = g_strdup_printf("Default (%s)", encoding);
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , sEncodingDefault, "Chinese", "Greek", "Japanese (SHIFT_JIS)"
      , "Korean", "Russian", "Unicode (UTF-8)", "Western (ISO-8859-1)", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 0); //...
     gtk_widget_set_sensitive(combobox, FALSE); //...
     FILLED_ADD(combobox, 1, 2, 3, 4);
     button = gtk_button_new_with_label("Advanced settings");
     gtk_widget_set_sensitive(button, FALSE); //...
     WIDGET_ADD(button, 1, 2, 3, 4);
     FRAME_NEW("Default colors");
     TABLE_NEW(2, 4);
     SEMI_INDENTED_ADD(gtk_label_new("Text color"), 0, 1, 0, 1);
     colorbutton = gtk_color_button_new();
     gtk_widget_set_sensitive(colorbutton, FALSE); //...
     WIDGET_ADD(colorbutton, 1, 2, 0, 1);
     SEMI_INDENTED_ADD(gtk_label_new("Background color"), 2, 3, 0, 1);
     colorbutton = gtk_color_button_new();
     gtk_widget_set_sensitive(colorbutton, FALSE); //...
     WIDGET_ADD(colorbutton, 3, 4, 0, 1);
     SEMI_INDENTED_ADD(gtk_label_new("Normal link color"), 0, 1, 1, 2);
     colorbutton = gtk_color_button_new();
     gtk_widget_set_sensitive(colorbutton, FALSE); //...
     WIDGET_ADD(colorbutton, 1, 2, 1, 2);
     SEMI_INDENTED_ADD(gtk_label_new("Visited link color"), 2, 3, 1, 2);
     colorbutton = gtk_color_button_new();
     gtk_widget_set_sensitive(colorbutton, FALSE); //...
     WIDGET_ADD(colorbutton, 3, 4, 1, 2);
     // Page "Interface"
     PAGE_NEW("Interface");
     FRAME_NEW("Navigationbar");
     TABLE_NEW(5, 2);
     INDENTED_ADD(gtk_label_new_with_mnemonic("_Toolbar style"), 0, 1, 0, 1);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "Default", "Icons", "Text", "Both", "Both horizontal", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), config->iToolbarStyle);
     g_signal_connect(combobox, "changed"
     , G_CALLBACK(on_settings_toolbarstyle_changed), browser);
     FILLED_ADD(combobox, 1, 2, 0, 1);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Use small icons in the toolbar");
     gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton)
      , config->iToolbarIconSize > 0);
     g_signal_connect(checkbutton, "toggled"
     , G_CALLBACK(on_settings_toolbariconsize_toggled), browser);
     SPANNED_ADD(checkbutton, 0, 2, 1, 2);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Show new tab button in the toolbar");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 2, 2, 3);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Show closed tabs button in the toolbar");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 2, 3, 4);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Show web search in the toolbar");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 2, 4, 5);
     FRAME_NEW("Miscellaneous");
     TABLE_NEW(3, 2);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Show close buttons on tabs");
     gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton)
      , config->iTabClose);
     g_signal_connect(checkbutton, "toggled"
     , G_CALLBACK(on_settings_closebuttonsontabs_toggled), browser);
     SPANNED_ADD(checkbutton, 0, 2, 0, 1);
     INDENTED_ADD(gtk_label_new_with_mnemonic("Tabbar _position"), 0, 1, 1, 2);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "Left", "Top", "Right", "Bottom", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 1); //...
     gtk_widget_set_sensitive(combobox, FALSE); //...
     FILLED_ADD(combobox, 1, 2, 1, 2);
     INDENTED_ADD(gtk_label_new("Location search engine"), 0, 1, 2, 3);
     entry = gtk_entry_new();
     gtk_entry_set_text(GTK_ENTRY(entry), config->sLocationSearch);
     g_signal_connect(entry, "focus-out-event"
     , G_CALLBACK(on_settings_locationsearch_focus_out), browser);
     FILLED_ADD(entry, 1, 2, 2, 3);
     // Page "Network"
     PAGE_NEW("Network");
     FRAME_NEW("Proxy Server");
     gtk_widget_set_sensitive(frame, FALSE); //...
     TABLE_NEW(8, 3);
     GtkWidget* radiobutton = radio_button_new(NULL, "_Automatically");
     SPANNED_ADD(radiobutton, 0, 1, 0, 1);
     SPANNED_ADD(radio_button_new(GTK_RADIO_BUTTON(radiobutton), "_No proxy server"), 0, 1, 1, 2);
     SPANNED_ADD(radio_button_new(GTK_RADIO_BUTTON(radiobutton), "_Manual settings"), 0, 1, 2, 3);
     INDENTED_ADD(gtk_label_new("Host"), 0, 1, 3, 4);
     FILLED_ADD(gtk_entry_new(), 1, 2, 3, 4);
     INDENTED_ADD(gtk_label_new("Port"), 0, 1, 4, 5);
     FILLED_ADD(gtk_entry_new(), 1, 2, 4, 5);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Proxy requires authentication");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     // TODO The proxy user and pass need to be indented further
     SPANNED_ADD(checkbutton, 0, 2, 5, 6);
     INDENTED_ADD(gtk_label_new("Username"), 0, 1, 6, 7);
     FILLED_ADD(gtk_entry_new(), 1, 2, 6, 7);
     INDENTED_ADD(gtk_label_new("Password"), 0, 1, 7, 8);
     FILLED_ADD(gtk_entry_new(), 1, 2, 7, 8);
     FRAME_NEW("Cache");
     TABLE_NEW(2, 2);
     INDENTED_ADD(gtk_label_new("Cache size"), 0, 1, 0, 1);
     hbox = gtk_hbox_new(FALSE, 4);
     spinbutton = gtk_spin_button_new_with_range(0, 10000, 10);
     gtk_spin_button_set_value(GTK_SPIN_BUTTON(spinbutton), 100/*config->iCacheSize*/);
     gtk_widget_set_sensitive(spinbutton, FALSE); //...
     gtk_box_pack_start(GTK_BOX(hbox), spinbutton, FALSE, FALSE, 0);
     gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("MB"), FALSE, FALSE, 0);
     SPANNED_ADD(hbox, 1, 2, 0, 1);
     button = gtk_button_new_with_label("Clear cache");
     gtk_widget_set_sensitive(button, FALSE); //...
     SPANNED_ADD(button, 2, 3, 0, 1);
     // Page "Privacy"
     PAGE_NEW("Privacy");
     FRAME_NEW("Cookies");
     TABLE_NEW(4, 2);
     INDENTED_ADD(gtk_label_new("Accept cookies"), 0, 1, 0, 1);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "All cookies", "Session cookies", "None", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 0); //...
     gtk_widget_set_sensitive(combobox, FALSE); //...
     FILLED_ADD(combobox, 1, 2, 0, 1);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Allow cookies from the original website only");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 2, 1, 2);
     INDENTED_ADD(gtk_label_new("Maximum cookie age"), 0, 1, 2, 3);
     hbox = gtk_hbox_new(FALSE, 4);
     spinbutton = gtk_spin_button_new_with_range(0, 360, 1);
     gtk_spin_button_set_value(GTK_SPIN_BUTTON(spinbutton), 30/*config->iCookieAgeMax*/);
     gtk_widget_set_sensitive(spinbutton, FALSE); //...
     gtk_box_pack_start(GTK_BOX(hbox), spinbutton, FALSE, FALSE, 0);
     gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("days"), FALSE, FALSE, 0);
     SPANNED_ADD(hbox, 1, 2, 2, 3);
     button = gtk_button_new_with_label("View cookies");
     gtk_widget_set_sensitive(button, FALSE); //...
     WIDGET_ADD(button, 1, 2, 3, 4);
     // Page "Programs"
     // TODO: We want to see program icons
     // TODO: Do we want to see available programs only?
     // TODO: Is hardcoding programs a good idea?
     // TODO: We should try to get programs from a global place
     // TODO: We need to determine 'default' programs somehow
     // TODO: Should we add an edonkey client?
     PAGE_NEW("Programs");
     FRAME_NEW("External programs");
     TABLE_NEW(5, 2);
     INDENTED_ADD(gtk_label_new("Text editor"), 0, 1, 0, 1);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "Internal", "Default", "geany", "gedit", "medit", "mousepad"
       , "scite", "Other...", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 0); //...
     gtk_widget_set_sensitive(combobox, FALSE); //...
     FILLED_ADD(combobox, 1, 2, 0, 1);
     INDENTED_ADD(gtk_label_new("Email client"), 0, 1, 1, 2);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "Default", "Balsa", "Claws Mail", "Evolution", "Thunderbird"
      , "Other...", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 0); //...
     gtk_widget_set_sensitive(combobox, FALSE); //...
     FILLED_ADD(combobox, 1, 2, 1, 2);
     INDENTED_ADD(gtk_label_new("Download manager"), 0, 1, 2, 3);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "None", "Gwget", "Urlgfe", "Other...", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 0); //...
     gtk_widget_set_sensitive(combobox, FALSE); //...
     FILLED_ADD(combobox, 1, 2, 2, 3);
     INDENTED_ADD(gtk_label_new("BitTorrent client"), 0, 1, 3, 4);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "None", "Azureus", "BitTorrent", "Deluge", "Transmission"
      , "Other...", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 0); //...
     gtk_widget_set_sensitive(combobox, FALSE); //...
     FILLED_ADD(combobox, 1, 2, 3, 4);
     INDENTED_ADD(gtk_label_new("Feed reader"), 0, 1, 4, 5);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "None", "Akregator", "Blam!", "RSSOwl", "Straw", "Other...", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 0); //...
     gtk_widget_set_sensitive(combobox, FALSE); //...
     FILLED_ADD(combobox, 1, 2, 4, 5);
     gtk_box_pack_start(GTK_BOX(GTK_DIALOG(settings_dialog)->vbox)
      , notebook, FALSE, FALSE, 0);
     gtk_widget_show_all(settings_dialog);
    }
}

void on_action_toolbar_navigation_activate(GtkToggleAction* action, CBrowser* browser)
{
    config->bToolbarNavigation = gtk_toggle_action_get_active(action);
    sokoke_widget_set_visible(browser->navibar, config->bToolbarNavigation);
}

void on_action_toolbar_bookmarks_activate(GtkToggleAction* action, CBrowser* browser)
{
    config->bToolbarBookmarks = gtk_toggle_action_get_active(action);
    sokoke_widget_set_visible(browser->bookmarkbar, config->bToolbarBookmarks);
}

void on_action_toolbar_downloads_activate(GtkToggleAction* action, CBrowser* browser)
{
    /*config->bToolbarDownloads = gtk_toggle_action_get_active(action);
    sokoke_widget_set_visible(browser->downloadbar, config->bToolbarDownloads);*/
}

void on_action_toolbar_status_activate(GtkToggleAction* action, CBrowser* browser)
{
    config->bToolbarStatus = gtk_toggle_action_get_active(action);
    sokoke_widget_set_visible(browser->statusbar, config->bToolbarStatus);
}

void on_action_refresh_activate(GtkAction* action, CBrowser* browser)
{
    GdkModifierType state = (GdkModifierType)0;
    gint x, y; gdk_window_get_pointer(NULL, &x, &y, &state);
    document_refresh(get_current_document(browser), state & GDK_SHIFT_MASK);
}

void on_action_refresh_stop_activate(GtkAction* action, CBrowser* browser)
{
    gchar* sStockId; g_object_get(action, "stock-id", &sStockId, NULL);
    // Refresh or stop, depending on the stock id
    if(!strcmp(sStockId, GTK_STOCK_REFRESH))
    {
     GdkModifierType state = (GdkModifierType)0;
     gint x, y; gdk_window_get_pointer(NULL, &x, &y, &state);
     document_refresh(get_current_document(browser), state & GDK_SHIFT_MASK);
    }
    else
     document_stop(get_current_document(browser));
    g_free(sStockId);
}

void on_action_stop_activate(GtkAction* action, CBrowser* browser)
{
    document_stop(get_current_document(browser));
}

void on_action_zoom_in_activate(GtkAction* action, CBrowser* browser)
{
    const gfloat zoom = document_get_text_size(get_current_document(browser));
    document_set_text_size(get_current_document(browser), zoom + 0.1);
}

void on_action_zoom_out_activate(GtkAction* action, CBrowser* browser)
{
    const gfloat zoom = document_get_text_size(get_current_document(browser));
    document_set_text_size(get_current_document(browser), zoom - 0.1);
}

void on_action_zoom_normal_activate(GtkAction* action, CBrowser* browser)
{
    document_set_text_size(get_current_document(browser), 1);
}

void on_action_back_activate(GtkAction* action, CBrowser* browser)
{
    document_go_back(get_current_document(browser));
}

void on_action_forward_activate(GtkAction* action, CBrowser* browser)
{
    document_go_forward(get_current_document(browser));
}

void on_action_home_activate(GtkAction* action, CBrowser* browser)
{
    document_load_uri(get_current_document(browser), config->sHomepage);
}

void on_menu_tabsclosed_item_activate(GtkMenuItem* menuitem, CBrowser* browser)
{
    DEBUG_FUNC
    // Create a new document with an uri which has been closed before
    const gchar* sUri = (const gchar*)g_object_get_data(
     G_OBJECT(menuitem), "uri");
    CBrowser* curBrowser = browser_new(browser);
    document_load_uri(curBrowser->document, sUri);
    gtk_widget_destroy(GTK_WIDGET(menuitem));
    update_browser_actions(curBrowser);
}

void on_action_tabclosed_undo_activate(GtkAction* action, CBrowser* browser)
{
    DEBUG_FUNC
    // Emulate a click on the most recent closedtabs item
    GtkWidget* menuitem;
    GList* children = gtk_container_get_children(
     GTK_CONTAINER(browser->menu_tabsclosed));
    menuitem = (GtkWidget*)g_list_nth_data(children, 0);
    on_menu_tabsclosed_item_activate(GTK_MENU_ITEM(menuitem), browser);
    g_list_free(children);
}

void on_action_tabsclosed_clear_activate(GtkAction* action, CBrowser* browser)
{
    DEBUG_FUNC
    // Clear the closed tabs list, ie. all items that have an uri
    GtkWidget* menuitem; GtkWidget* icon;
    GList* children = gtk_container_get_children(
     GTK_CONTAINER(browser->menu_tabsclosed));
    do
    {
     menuitem = (GtkWidget*)children->data;
     if(g_object_get_data(G_OBJECT(menuitem), "uri"))
      gtk_widget_destroy(menuitem);
    }
    while(children = g_list_next(children)); g_list_free(children);
    update_browser_actions(browser);
}

void on_action_colorpicker_activate(GtkAction* action, CBrowser* browser)
{
    DEBUG_FUNC
    // Show the color picker. Create it if necessary.
    if(GTK_IS_DIALOG(color_picker))
     gtk_window_present(GTK_WINDOW(color_picker));
    else
    {
     color_picker
     = gtk_dialog_new_with_buttons("Color picker"
                                 , GTK_WINDOW(browser->window)
                                 , GTK_DIALOG_DESTROY_WITH_PARENT
                                 , GTK_STOCK_CLOSE
                                 , GTK_RESPONSE_CLOSE
                                 , NULL);
     gtk_window_set_icon_name(GTK_WINDOW(color_picker), GTK_STOCK_COLOR_PICKER);
     gtk_dialog_set_has_separator(GTK_DIALOG(color_picker), TRUE);
     g_signal_connect(GTK_OBJECT(color_picker), "response"
      , GTK_SIGNAL_FUNC(gtk_widget_destroy), color_picker);
     gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(color_picker)->vbox), 5);
     GtkWidget* color_selection = gtk_color_selection_new();
     gtk_container_set_border_width(GTK_CONTAINER(color_selection), 5);
     // TODO: Could we do anything to make the opacity useful?
     gtk_color_selection_set_has_opacity_control(
      GTK_COLOR_SELECTION(color_selection), FALSE);
     // TODO: Is the palette any good?
     gtk_color_selection_set_has_palette(
      GTK_COLOR_SELECTION(color_selection), TRUE);
     gtk_box_pack_start(GTK_BOX(GTK_DIALOG(color_picker)->vbox)
      , color_selection, TRUE, TRUE, 15);
     gtk_box_set_spacing(GTK_BOX(GTK_DIALOG(color_picker)->vbox), 2);
     // TODO: Any useful buttons we could provide?
     gtk_widget_show_all(color_picker);
    }
}

void on_action_tab_previous_activate(GtkAction* action, CBrowser* browser)
{
    gint page = gtk_notebook_get_current_page(GTK_NOTEBOOK(browser->documents));
    gtk_notebook_set_current_page(GTK_NOTEBOOK(browser->documents), page - 1);
}

void on_action_tab_next_activate(GtkAction* action, CBrowser* browser)
{
    // Advance one tab or jump to the first one if we are at the last one
    gint page = gtk_notebook_get_current_page(GTK_NOTEBOOK(browser->documents));
    if(page == gtk_notebook_get_n_pages(GTK_NOTEBOOK(browser->documents)) - 1)
     page = -1;
    gtk_notebook_set_current_page(GTK_NOTEBOOK(browser->documents), page + 1);
}

void on_window_menu_item_activate(GtkImageMenuItem* widget, CBrowser* browser)
{
    gint page = get_document_index(browser->document, browser);
    gtk_notebook_set_current_page(GTK_NOTEBOOK(browser->documents), page);
}

void on_action_about_activate(GtkAction* action, CBrowser* browser)
{
    DEBUG_FUNC
    gtk_show_about_dialog(GTK_WINDOW(browser->window)
     , "logo-icon-name", "web-browser"
     , "name", PACKAGE_NAME
     , "version", PACKAGE_VERSION
     , "comments", "A lightweight web browser."
     , "copyright", "Copyright © 2007 Christian Dywan"
     , "website", "http://software.twotoasts.de"
     , "authors", credits_authors
     , "documenters", credits_documenters
     , "artists", credits_artists
     , "license", license_gpl_2
     , "wrap-license", TRUE
     //, "translator-credits", _("translator-credits")
     , NULL);
}

gboolean on_navibar_entry_location_key_down(GtkWidget* widget
 , GdkEventKey* event, CBrowser* browser)
{
    DEBUG_FUNC
    const gchar* sUri; CBrowser* curBrowser;
    switch(event->keyval)
    {
     case GDK_Return:
      sUri = gtk_entry_get_text(GTK_ENTRY(widget));
      if(sUri)
      {
       gchar* sNewUri = magic_uri(sUri);
       // TODO: Use sUrl as soon as completion is better
       // TODO Question: Shouldn't we autogenerate the completion from history?
       entry_completion_append(GTK_ENTRY(widget), sUri);
       document_load_uri(get_current_document(browser), sNewUri);
       g_free(sNewUri);
      }
      return TRUE;
     case GDK_Escape:
      // This is a workaround until document_get_uri works
      curBrowser = document_get_browser(get_current_document(browser));
      if(sUri = curBrowser->sUri)
      //if(sUri = document_get_uri(get_current_document(browser)))
       if(strlen(sUri)) gtk_entry_set_text(GTK_ENTRY(widget), sUri);
      return TRUE;
    }
    return FALSE;
}

void on_navibar_entry_search_activate(GtkWidget* widget, CBrowser* browser)
{
    DEBUG_FUNC
    // TODO: use custom search engine, hardcode google for now
    const gchar* sKeywords = gtk_entry_get_text(GTK_ENTRY(widget));
    gchar* sText = g_strconcat("http://www.google.com/search?q=", sKeywords, NULL);
    entry_completion_append(GTK_ENTRY(widget), sKeywords);
    document_load_uri(get_current_document(browser), sText);
    g_free(sText);
}

void on_navibar_tabsclosed_clicked(GtkToolItem* widget, CBrowser* browser)
{
    DEBUG_FUNC
    gtk_menu_popup(GTK_MENU(browser->menu_tabsclosed), NULL, NULL, NULL, NULL, 0
     , gtk_get_current_event_time());
}

void on_action_panels_activate(GtkToggleAction* action, CBrowser* browser)
{
    config->bPanels = gtk_toggle_action_get_active(action);
    sokoke_widget_set_visible(browser->panels, config->bPanels);
}

void on_action_panel_custom_activate(GtkToggleAction* action, CBrowser* browser)
{
    // TODO: Show panel if not already visible
    gboolean bActive = gtk_toggle_action_get_active(action);
    sokoke_widget_set_visible(gtk_widget_get_parent(
     browser->panels_notebook), bActive);
    gtk_paned_set_position(GTK_PANED(gtk_widget_get_parent(browser->panels))
     , gtk_toggle_action_get_active(action) ? config->iPanelPos : 0);
}

// FIXME: Is this the right place for this?
void on_action_openinpanel_activate(GtkToggleAction* action, CBrowser* browser)
{
    GtkWidget* document = get_current_document(browser);
    CBrowser* curBrowser = document_get_browser(document);
    g_free(config->sPanelCustom);
    config->sPanelCustom = g_strdup(curBrowser->sUri);
    document_load_uri(browser->panel_custom, curBrowser->sUri);
}

void on_panels_button_close_clicked(GtkWidget* widget, CBrowser* browser)
{
    config->bPanels = FALSE;
    gtk_widget_hide(browser->panels);
}

gint on_notebook_tab_mouse_down(GtkWidget* widget, GdkEventButton* event
 , CBrowser* browser)
{
    DEBUG_FUNC
    if(event->button == 1 && event->type == GDK_2BUTTON_PRESS)
    {
     // Toggle the label visibility on double click
     GtkWidget* child = gtk_bin_get_child(GTK_BIN(widget));
     GList* children = gtk_container_get_children(GTK_CONTAINER(child));
     child = (GtkWidget*)g_list_nth_data(children, 1);
     gboolean visible = gtk_widget_get_child_visible(GTK_WIDGET(child));
     gtk_widget_set_child_visible(GTK_WIDGET(child), !visible);
     gint a, b; sokoke_widget_get_text_size(browser->document_name, "M", &a, &b);
     gtk_widget_set_size_request(child, !visible
      ? a * config->iTabSize : 0, !visible ? -1 : 0);
     g_list_free(children);
     return TRUE;
    }
    else if(event->button == 2)
    {
     // Close the document on middle click
     document_close(browser->document, browser);
     return TRUE;
    }

    return FALSE;
}

gint on_notebook_tab_close_mouse_down(GtkWidget* widget, GdkEventButton* event
 , CBrowser* browser)
{
    DEBUG_FUNC
    document_close(browser->document, browser);

    return FALSE;
}

void on_notebook_switch_page(GtkWidget* widget, GtkNotebookPage* page
 , guint page_num, CBrowser* browser)
{
    DEBUG_FUNC
    GtkWidget* docbox = gtk_notebook_get_nth_page(
     GTK_NOTEBOOK(browser->documents), page_num);
    GtkWidget* document = gtk_bin_get_child(GTK_BIN(docbox));
    // This is a workaround until document_get_uri works
    CBrowser* curBrowser = document_get_browser(get_current_document(browser));
    const gchar* sUri = curBrowser->sUri;
    //const gchar* sUri = document_get_uri(document);
    if(sUri) gtk_entry_set_text(GTK_ENTRY(browser->navibar_location), sUri);
    // This is a workaround until document_get_title works
    const gchar* sTitle = curBrowser->sTitle;
    //const gchar* sTitle = document_get_title(document);
    if(sTitle)
    {
     gchar* sWindowTitle = g_strconcat(sTitle, " - "
      , PACKAGE_NAME, NULL);
     gtk_window_set_title(GTK_WINDOW(browser->window), sWindowTitle);
     g_free(sWindowTitle);
    }
    update_favicon(browser);
    update_security(browser);
    update_gui_state(browser);
    update_statusbar_text(browser);
    update_feeds(browser);
    update_search_engines(browser);
}

gboolean on_document_mouse_down(GtkWidget* widget, GdkEventButton* event
 , CBrowser* browser)
{
    DEBUG_FUNC
    GdkModifierType state = (GdkModifierType)0;
    gint x, y;
    gdk_window_get_pointer(NULL, &x, &y, &state);
    switch(event->button)
    {
     case 1:
      if(state & GDK_SHIFT_MASK)
      {
       // TODO: Open link in new window
       UNIMPLEMENTED return TRUE;
      }
      else if(state & GDK_MOD1_MASK)
      {
       // TODO: Open link in new tab
       UNIMPLEMENTED return TRUE;
      }
      break;
     case 2:
      if(state & GDK_CONTROL_MASK)
      {
       document_set_text_size(get_current_document(browser), 1);
       return TRUE;
      }
      else
      {
       // TODO: Middle click opens new tab
       UNIMPLEMENTED return TRUE;
      }
      break;
     case 3:
      on_document_popup(get_current_document(browser), event, browser);
      return TRUE;
    }
    return FALSE;
}

void on_document_popup(GtkWidget* document, GdkEventButton* event
 , CBrowser* browser)
{
    DEBUG_FUNC
    gboolean bLink = FALSE; // TODO: Did we right-click a link?
    gboolean bImage = FALSE; // TODO: Did we right-click an image?
    gchar* sSelection = NULL; // TODO: The possibly selected text

    action_set_visible("Back", !bLink && !bImage && !sSelection, browser);
    action_set_visible("Forward", !bLink && !bImage && !sSelection, browser);
    action_set_visible("Refresh", !bLink && !bImage && !sSelection, browser);
    action_set_visible("Stop", !bLink && !bImage && !sSelection, browser);
    action_set_visible("RefreshEvery", !bLink && !bImage && !sSelection, browser);

    action_set_visible("LinkTabNew", bLink, browser);
    action_set_visible("LinkTabCurrent", bLink, browser);
    action_set_visible("LinkWindowNew", bLink, browser);

    action_set_visible("LinkBookmarkNew", bLink, browser);
    action_set_visible("LinkSaveAs", bLink, browser);
    action_set_visible("LinkSaveWith", bLink, browser);
    action_set_visible("LinkCopy", bLink, browser);
    action_set_visible("LinkSend", bLink, browser);

    action_set_visible("SelectionLinksNewTabs", sSelection != NULL, browser);
    action_set_visible("SelectionTextTabNew", sSelection != NULL, browser);
    action_set_visible("SelectionTextTabCurrent", sSelection != NULL, browser);
    action_set_visible("SelectionTextWindowNew", sSelection != NULL, browser);

    action_set_visible("ImageViewTabNew", bImage, browser);
    action_set_visible("ImageViewTabCurrent", bImage, browser);
    action_set_visible("ImageSaveAs", bImage, browser);
    action_set_visible("ImageSaveWith", bImage, browser);
    action_set_visible("ImageCopy", bImage, browser);
    action_set_visible("ImageSend", bImage, browser);

    action_set_visible("UndoTabClose", !bLink && !bImage && !sSelection, browser);
    action_set_visible("BookmarkNew", !bLink && !bImage && !sSelection, browser);
    action_set_visible("SaveAs", !bLink && !bImage && !sSelection, browser);
    action_set_visible("Print", !bLink && !bImage && !sSelection, browser);

    action_set_visible("Copy", sSelection != NULL, browser);
    action_set_visible("SelectAll", sSelection != NULL, browser);
    action_set_visible("SelectionSearch", sSelection != NULL, browser);
    action_set_visible("SelectionSearchWith", sSelection != NULL, browser);
    action_set_visible("SelectionSourceView", sSelection != NULL, browser);

    action_set_visible("BackgroundImage", !bLink && !bImage && !sSelection, browser);
    action_set_visible("SourceView", !bLink && !bImage && !sSelection, browser);
    action_set_visible("Properties", !sSelection, browser);

    int button, event_time;
    if(event)
    {
     button = event->button;
     event_time = event->time;
    }
    else
    {
     button = 0;
     event_time = gtk_get_current_event_time();
    }

    gtk_menu_popup(GTK_MENU(browser->menu_document), NULL, NULL, NULL, NULL, button, event_time);
}

gboolean on_document_scroll(GtkWidget* document, GdkEventScroll* event
 , CBrowser* browser)
{
    DEBUG_FUNC
    GdkModifierType state = (GdkModifierType)0;
    gint x, y;
    gdk_window_get_pointer(NULL, &x, &y, &state);
    if(state & GDK_CONTROL_MASK)
    {
     const gfloat size = document_get_text_size(get_current_document(browser));
     if(event->direction == GDK_SCROLL_DOWN)
      document_set_text_size(get_current_document(browser), size + 0.1);
     else if(event->direction == GDK_SCROLL_UP)
      document_set_text_size(get_current_document(browser), size - 0.1);
     return TRUE;
    }
    else
     return FALSE;
}

void on_document_destroy(GtkWidget* widget, CBrowser* browser)
{
    DEBUG_FUNC
    // Update browser list, free memory and possibly quit
    browser_count--;
    GList* tmp_list = g_list_find(browser_list, browser);
    browser_list = g_list_remove_link(browser_list, tmp_list);
    g_free(browser->sUri);
    g_free(browser->sTitle);
    g_free(browser->sStatusMessage);
    if(!browser_count)
    {
     g_object_unref(browser->actiongroup);
     g_object_unref(browser->menu_document);
     gtk_main_quit();
    }
}

void on_findbox_button_close_clicked(GtkWidget* widget, CBrowser* browser)
{
    gtk_widget_hide(browser->findbox);
}

gboolean on_window_destroy(GtkWidget* widget, GdkEvent* event, CBrowser* browser)
{
    DEBUG_FUNC
    gboolean bProceed = TRUE;
    // TODO: What if there are multiple windows?
    // TODO: Smart dialog, à la 'Session?: Save, Discard, Cancel'
    // TODO: Pref startup: session, ask, homepage, blank <-- ask
    // TODO: Pref quit: session, ask, none <-- ask

    if(0 /*browser_count > 1*/)
    {
     GtkDialog* dialog;
     dialog = GTK_DIALOG(gtk_message_dialog_new(GTK_WINDOW(browser->window)
      , GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_WARNING, GTK_BUTTONS_YES_NO
      , "There is more than one tab open. Do you want to close anyway?"));
     gtk_window_set_title(GTK_WINDOW(dialog), PACKAGE_NAME);

     gtk_dialog_set_default_response(dialog, GTK_RESPONSE_YES);
     bProceed = gtk_dialog_run(dialog) == GTK_RESPONSE_YES;

     gtk_widget_destroy(GTK_WIDGET(dialog));
    }
    return !bProceed;
}

// -- Browser creation begins here

// FIXME: This needs to move, in case we still want it
GtkWidget* sokoke_frame_new_child_shadow(GtkWidget* child, GtkShadowType shadow)
{
    GtkWidget* frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), shadow);
    gtk_container_add(GTK_CONTAINER(frame), child);
    return frame;
}

// FIXME: This does not belong here
// TODO: Investigate what signature notify::position actually has
gboolean on_panels_notify_position_activate(GtkPaned* paned, CBrowser* browser)
{
    guint pos = gtk_paned_get_position(paned);
    //if(!pos) g_object_set_data(G_OBJECT(paned), "position-null", TRUE);
    config->iPanelPos = pos;
    //g_print("%d %d\n", pos, config->iPanelPos);
    return FALSE;
}

static CBrowser* browser_new(CBrowser* oldBrowser)
{
    DEBUG_FUNC
    browser_count++;
    CBrowser* browser = g_new0(CBrowser, 1);
    browser_list = g_list_prepend(browser_list, browser);

    if(!oldBrowser) {

    GtkWidget* label; GtkWidget* hbox;

    // Setup the window
    browser->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    //sokoke_widget_set_transparency(browser->window, 0.7);
    GdkScreen* screen = gtk_window_get_screen(GTK_WINDOW(browser->window));
    gdouble width = gdk_screen_get_width(screen) / 1.7;
    gdouble height = gdk_screen_get_height(screen) / 1.7;
    gtk_window_set_default_size(GTK_WINDOW(browser->window)
     , (gint)width, (gint)height);
    // FIXME: Use custom program icon
    gtk_window_set_icon_name(GTK_WINDOW(browser->window), "web-browser");
    gtk_window_set_title(GTK_WINDOW(browser->window), PACKAGE_NAME);
    gtk_window_add_accel_group(GTK_WINDOW(browser->window), accel_group);
    g_signal_connect(GTK_OBJECT(browser->window), "delete-event"
     , G_CALLBACK(on_window_destroy), browser);
    GtkWidget* vbox = gtk_vbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER(browser->window), vbox);
    gtk_widget_show(vbox);

    // Actions
    // TODO: Fill in a good description for each 'hm?'
    // TODO: We need more mnemonics
    static GtkActionEntry entries[] = {
     { "File", NULL, "_File" },
     { "WindowNew", STOCK_WINDOW_NEW
     , NULL, "<Ctrl>n"
     , "Open a new window", G_CALLBACK(on_action_window_new_activate) },
     { "TabNew", STOCK_TAB_NEW
     , NULL, "<Ctrl>t"
     , "Open a new tab", G_CALLBACK(on_action_tab_new_activate) },
     { "Open", GTK_STOCK_OPEN
     , NULL, "<Ctrl>o"
     , "Open a file", NULL/*G_CALLBACK(on_action_open_activate)*/ },
     { "SaveAs", GTK_STOCK_SAVE_AS
     , NULL, "<Ctrl>s"
     , "Save to a file", NULL/*G_CALLBACK(on_action_saveas_activate)*/ },
     // TODO: Maybe we need a better label here
     { "Send", STOCK_SEND
     , "Send via email", ""
     , "Send with the chosen email client", NULL/*G_CALLBACK(on_action_send_activate)*/ },
     { "TabClose", STOCK_TAB_CLOSE
     , NULL, "<Ctrl>w"
     , "Close the current tab", G_CALLBACK(on_action_tab_close_activate) },
     { "WindowClose", STOCK_WINDOW_CLOSE
     , NULL, "<Ctrl><Shift>w"
     , "Close this window", G_CALLBACK(on_action_window_close_activate) },
     { "PageSetup", GTK_STOCK_PROPERTIES
     , "Page setup", ""
     , "hm?", NULL/*G_CALLBACK(on_action_page_setup_activate)*/ },
     { "PrintPreview", GTK_STOCK_PRINT_PREVIEW
     , NULL, ""
     , "hm?", NULL/*G_CALLBACK(on_action_print_preview_activate)*/ },
     { "Print", GTK_STOCK_PRINT
     , NULL, "<Ctrl>p"
     , "hm?", NULL/*G_CALLBACK(on_action_print_activate)*/ },
     { "Quit", GTK_STOCK_QUIT
     , NULL, "<Ctrl>q"
     , "Quit the application", G_CALLBACK(on_action_quit_activate) },
 
     { "Edit", NULL, "_Edit" },
     { "Undo", GTK_STOCK_UNDO
     , NULL, "<Ctrl>z"
     , "Undo the last modification", NULL/*G_CALLBACK(on_action_undo_activate)*/ },
     { "Redo", GTK_STOCK_REDO
     , NULL, "<Ctrl><Shift>z"
     , "Redo the last modification", NULL/*G_CALLBACK(on_action_redo_activate)*/ },
     { "Cut", GTK_STOCK_CUT
     , NULL, "<Ctrl>x"
     , "Cut the selected text", NULL/*G_CALLBACK(on_action_cut_activate)*/ },
     { "Copy", GTK_STOCK_COPY
     , NULL, "<Ctrl>c"
     , "Copy the selected text", NULL/*G_CALLBACK(on_action_copy_activate)*/ },
     { "Paste", GTK_STOCK_PASTE
     , NULL, "<Ctrl>v"
     , "Paste text from the clipboard", NULL/*G_CALLBACK(on_action_paste_activate)*/ },
     { "Delete", GTK_STOCK_DELETE
     , NULL, "Delete"
     , "Delete the selected text", NULL/*G_CALLBACK(on_action_delete_activate)*/ },
     { "SelectAll", GTK_STOCK_SELECT_ALL
     , NULL, "<Ctrl>a"
     , "Selected all text", NULL/*G_CALLBACK(on_action_selectall_activate)*/ },
     { "FormFill", STOCK_FORM_FILL
     , NULL, ""
     , "hm?", NULL/*G_CALLBACK(on_action_formfill_activate)*/ },
     { "Find", GTK_STOCK_FIND
     , NULL, "<Ctrl>f"
     , "hm?", G_CALLBACK(on_action_find_activate) },
     { "FindNext", GTK_STOCK_GO_FORWARD
     , "Find next", "<Ctrl>g"
     , "hm?", G_CALLBACK(on_action_find_next_activate) },
     { "FindPrevious", GTK_STOCK_GO_BACK
     , "Find previous", "<Ctrl><Shift>g"
     , "hm?", G_CALLBACK(on_action_find_previous_activate) },
     { "FindQuick", GTK_STOCK_FIND
     , "Quick find", "period"
     , "hm?", NULL/*G_CALLBACK(on_action_find_quick_activate)*/ },
     { "Preferences", GTK_STOCK_PREFERENCES
     , NULL, ""
     , "hm?", G_CALLBACK(on_action_preferences_activate) },

     { "View", NULL, "_View" },
     { "Refresh", GTK_STOCK_REFRESH
     , NULL, "<Ctrl>r"
     , "Refresh the current page", G_CALLBACK(on_action_refresh_stop_activate) },
     // TODO: Is appointment-new a good choice?
     // TODO: What if it isn't available?
     { "RefreshEvery", "appointment-new"
     , "Refresh every...", ""
     , "Refresh the current page", G_CALLBACK(on_action_refresh_stop_activate) },
     { "Stop", GTK_STOCK_STOP
     , NULL, "Escape"
     , "Stop loading of the current page", G_CALLBACK(on_action_refresh_stop_activate) },
     { "RefreshStop", GTK_STOCK_REFRESH
     , NULL, ""
     , NULL, G_CALLBACK(on_action_refresh_stop_activate) },
     { "ZoomIn", GTK_STOCK_ZOOM_IN
     , NULL, "<Ctrl>plus"
     , "hm?", G_CALLBACK(on_action_zoom_in_activate) },
     { "ZoomOut", GTK_STOCK_ZOOM_OUT
     , NULL, "<Ctrl>minus"
     , "hm?", G_CALLBACK(on_action_zoom_out_activate) },
     { "ZoomNormal", GTK_STOCK_ZOOM_100
     , NULL, "<Ctrl>0"
     , "hm?", G_CALLBACK(on_action_zoom_normal_activate) },
     { "BackgroundImage", STOCK_IMAGE
     , "Background image", ""
     , "hm?", NULL/*G_CALLBACK(on_action_background_image_activate)*/ },
     { "SourceView", STOCK_SOURCE_VIEW
     , NULL, ""
     , "hm?", NULL/*G_CALLBACK(on_action_source_view_activate)*/ },
     { "SelectionSourceView", STOCK_SOURCE_VIEW
     , "View selection source", ""
     , "hm?", NULL/*G_CALLBACK(on_action_selection_source_view_activate)*/ },
     { "Properties", GTK_STOCK_PROPERTIES
     , NULL, ""
     , "hm?", NULL/*G_CALLBACK(on_action_properties_activate)*/ },

     { "Go", NULL, "_Go" },
     { "Back", GTK_STOCK_GO_BACK
     , NULL, "<Alt>Left"
     , "hm?", G_CALLBACK(on_action_back_activate) },
     { "Forward", GTK_STOCK_GO_FORWARD
     , NULL, "<Alt>Right"
     , "hm?", G_CALLBACK(on_action_forward_activate) },
     { "Home", STOCK_HOMEPAGE
     , NULL, "<Alt>Home"
     , "hm?", G_CALLBACK(on_action_home_activate) },
     { "Location", GTK_STOCK_JUMP_TO
     , "Location...", "<Ctrl>l"
     , "hm?", NULL/*G_CALLBACK(on_action_go_location_activate)*/ },
     { "Websearch", GTK_STOCK_FIND
     , "Websearch...", "<Ctrl><Shift>f"
     , "hm?", NULL/*G_CALLBACK(on_action_websearch_activate)*/ },
     { "OpenInPanel", GTK_STOCK_JUMP_TO
     , "Open in panel...", ""
     , "hm?", G_CALLBACK(on_action_openinpanel_activate) },
     { "TabsClosed", STOCK_USER_TRASH
     , "Closed tabs", ""
     , "hm?", NULL },
     { "TabsClosedClear", GTK_STOCK_CLEAR
     , "Clear list of closed tabs", ""
     , "hm?", G_CALLBACK(on_action_tabsclosed_clear_activate) },
     { "UndoTabClose", GTK_STOCK_UNDELETE
     , "Undo close tab", ""
     , "hm?", G_CALLBACK(on_action_tabclosed_undo_activate) },
     { "LinkTabNew", STOCK_TAB_NEW
     , "Open link in new tab", ""
     , "hm?", NULL },
     { "LinkTabCurrent", NULL
     , "Open link in current tab", ""
     , "hm?", NULL },
     { "LinkWindowNew", STOCK_WINDOW_NEW
     , "Open link in new window", ""
     , "hm?", NULL },
     { "LinkBookmarkNew", STOCK_BOOKMARK_NEW
     , NULL, ""
     , "Bookmark this link", NULL/*G_CALLBACK(on_action_link_bookmark_activate)*/ },
     { "LinkSaveAs", GTK_STOCK_SAVE
     , "Save destination as...", ""
     , "Save destination to a file", NULL/*G_CALLBACK(on_action_link_saveas_activate)*/ },
     { "LinkSaveWith", STOCK_DOWNLOADS
     , "Save destination with download manager", ""
     , "Save destination with the chosen download manager", NULL/*G_CALLBACK(on_action_link_savewith_activate)*/ },
     { "LinkCopy", GTK_STOCK_COPY
     , "Copy link address", ""
     , "Copy the link address to the clipboard", NULL/*G_CALLBACK(on_action_link_copy_activate)*/ },
     { "LinkSend", STOCK_SEND
     , "Send link via email", ""
     , "Send link with the chosen email client", NULL/*G_CALLBACK(on_action_link_send_activate)*/ },
     { "SelectionLinksNewTabs", NULL
     , "Open selected links in tabs", ""
     , "hm?", NULL/*G_CALLBACK(on_action_properties_selection_activate)*/ },
     { "SelectionTextTabNew", STOCK_TAB_NEW
     , "Open <selection> in new tab", ""
     , "hm?", NULL/*G_CALLBACK(on_action_properties_selection_activate)*/ },
     { "SelectionTextTabCurrent", NULL
     , "Open <selection> in current tab", ""
     , "hm?", NULL/*G_CALLBACK(on_action_properties_selection_activate)*/ },
     { "SelectionTextWindowNew", STOCK_WINDOW_NEW
     , "Open <selection> in new window", ""
     , "hm?", NULL/*G_CALLBACK(on_action_properties_selection_activate)*/ },
     { "SelectionSearch", GTK_STOCK_FIND
     , "Search for <selection>", ""
     , "hm?", NULL/*G_CALLBACK(on_action_properties_selection_activate)*/ },
     { "SelectionSearchWith", GTK_STOCK_FIND
     , "Search for <selection> with...", ""
     , "hm?", NULL/*G_CALLBACK(on_action_properties_selection_activate)*/ },
     { "ImageViewTabNew", STOCK_TAB_NEW
     , "View image in new tab", ""
     , "hm?", NULL/*G_CALLBACK(on_action_properties_selection_activate)*/ },
     { "ImageViewTabCurrent", NULL
     , "View image in current tab", ""
     , "hm?", NULL/*G_CALLBACK(on_action_properties_selection_activate)*/ },
     { "ImageSaveAs", GTK_STOCK_SAVE
     , "Save image as...", ""
     , "Save image to a file", NULL/*G_CALLBACK(on_action_properties_selection_activate)*/ },
     { "ImageSaveWith", STOCK_DOWNLOADS
     , "Save image with download manager", ""
     , "Save image with the chosen download manager", NULL/*G_CALLBACK(on_action_properties_selection_activate)*/ },
     { "ImageCopy", GTK_STOCK_COPY
     , "Copy image address", ""
     , "Copy the image address to the clipboard", NULL/*G_CALLBACK(on_action_properties_selection_activate)*/ },
     { "ImageSend", STOCK_SEND
     , "Send image via email", ""
     , "Send image with the chosen email client", NULL/*G_CALLBACK(on_action_properties_selection_activate)*/ },

     { "Bookmarks", NULL, "_Bookmarks" },
     { "BookmarkNew", STOCK_BOOKMARK_NEW
     , NULL, "<Ctrl>d"
     , "hm?", NULL/*G_CALLBACK(on_action_bookmark_new_activate)*/ },
     { "BookmarksManage", STOCK_BOOKMARKS
     , "_Manage bookmarks", "<Ctrl>b"
     , "hm?", NULL/*G_CALLBACK(on_action_bookmarks_manage_activate)*/ },

     { "Tools", NULL, "_Tools" },
     { "ColorPicker", GTK_STOCK_COLOR_PICKER
     , "_Color picker", ""
     , "hm?", G_CALLBACK(on_action_colorpicker_activate) },

     { "Window", NULL, "_Window" },
     { "SessionLoad", NULL
     , "Load session", ""
     , "hm?", NULL/*G_CALLBACK(on_action_session_load_activate)*/ },
     { "SessionSave", NULL
     , "Save session", ""
     , "hm?", NULL/*G_CALLBACK(on_action_session_save_activate)*/ },
     { "TabPrevious", NULL
     , "Previous tab", "<Ctrl>Page_Up"
     , "hm?", G_CALLBACK(on_action_tab_previous_activate) },
     { "TabNext", NULL
     , "Next tab", "<Ctrl>Page_Down"
     , "hm?", G_CALLBACK(on_action_tab_next_activate) },
     { "TabOverview", NULL
     , "Tab overview", ""
     , "hm?", NULL/*G_CALLBACK(on_action_tab_overview_activate)*/ },

     { "Help", NULL, "_Help" },
     { "HelpContents", GTK_STOCK_HELP
     , "_Contents", "F1"
     , "hm?", NULL/*G_CALLBACK(on_action_help_contents_activate)*/ },
     { "About", GTK_STOCK_ABOUT
     , NULL, ""
     , "hm?", G_CALLBACK(on_action_about_activate) },
    };
    static guint entries_n = G_N_ELEMENTS(entries);

    static GtkToggleActionEntry toggle_entries[] = {
     { "PrivateBrowsing", NULL
     , "Private browsing", ""
     , "hm?", NULL/*G_CALLBACK(on_action_private_browsing_activate)*/
     , FALSE },
     { "WorkOffline", GTK_STOCK_DISCONNECT
     , "Work offline", ""
     , "hm?", NULL/*G_CALLBACK(on_action_work_offline_activate)*/
     , FALSE },

     { "Menubar", NULL
     , "_Menubar", ""
     , "hm?", NULL/*G_CALLBACK(on_action_menu_activate)*/
     , FALSE },
     { "ToolbarNavigation", NULL
     , "Navigationbar", ""
     , "hm?", G_CALLBACK(on_action_toolbar_navigation_activate)
     , FALSE },
     { "Panels", NULL
     , "_Panels", "F4"
     , "hm?", G_CALLBACK(on_action_panels_activate)
     , FALSE },
     { "ToolbarBookmarks", NULL
     , "Bookmarkbar", ""
     , "hm?", G_CALLBACK(on_action_toolbar_bookmarks_activate)
     , FALSE },
     { "ToolbarDownloads", NULL
     , "Downloadbar", ""
     , "hm?", NULL/*G_CALLBACK(on_action_toolbar_downloads_activate)*/
     , FALSE },
     { "ToolbarStatus", NULL
     , "Statusbar", ""
     , "hm?", G_CALLBACK(on_action_toolbar_status_activate)
     , FALSE },
     { "RefreshEveryEnable", NULL
     , "Enabled", ""
     , "hm?", NULL/*G_CALLBACK(on_action_reloadevery_enable_activate)*/
     , FALSE },

     { "PanelBookmarks", STOCK_BOOKMARKS
     , "_Bookmarks", ""
     , "hm?", NULL/*G_CALLBACK(on_action_panel_bookmarks_activate)*/
     , FALSE },
     { "PanelDownloads", STOCK_DOWNLOADS
     , NULL, ""
     , "hm?", NULL/*G_CALLBACK(on_action_panel_downloads_activate)*/
     , FALSE },
     { "PanelConsole", STOCK_CONSOLE
     , NULL, ""
     , "hm?", NULL/*G_CALLBACK(on_action_panel_errorconsole_activate)*/ },
     { "PanelExtensions", STOCK_EXTENSIONS
     , NULL, ""
     , "hm?", NULL/*G_CALLBACK(on_action_panel_extensions_activate)*/ },
     { "PanelHistory", STOCK_HISTORY
     , "_History", ""
     , "hm?", NULL/*G_CALLBACK(on_action_panel_history_activate)*/
     , FALSE },
     { "PanelTabs", NULL
     , "_Tabs", ""
     , "hm?", NULL/*G_CALLBACK(on_action_panel_tabs_activate)*/
     , FALSE },
     { "PanelCustom", GTK_STOCK_CONVERT
     , "_Custom", ""
     , "hm?", G_CALLBACK(on_action_panel_custom_activate)
     , FALSE },
     { "ReloadEveryActive", NULL
     , "_Active", ""
     , "hm?", NULL/*G_CALLBACK(on_action_reloadevery_active_activate)*/
     , FALSE },
    };
    static guint toggle_entries_n = G_N_ELEMENTS(toggle_entries);

    static GtkRadioActionEntry refreshevery_entries[] = {
     { "RefreshEvery30", NULL
     , "30 seconds", ""
     , "Refresh every 30 seconds", 30 },
     { "RefreshEvery60", NULL
     , "60 seconds", ""
     , "Refresh every 60 seconds", 60 },
     { "RefreshEvery300", NULL
     , "5 minutes", ""
     , "Refresh every 5 minutes", 300 },
     { "RefreshEvery900", NULL
     , "15 minutes", ""
     , "Refresh every 15 minutes", 900 },
     { "RefreshEvery1800", NULL
     , "30 minutes", ""
     , "Refresh every 30 minutes", 1800 },
     { "RefreshEveryCustom", NULL
     , "Custom...", ""
     , "Refresh by a custom period", 0 },
    };
    static guint refreshevery_entries_n = G_N_ELEMENTS(refreshevery_entries);

    // Layout for menubar, navibar and document popup
    static const gchar* ui_markup =
    "<ui>"
     "<menubar>"
      "<menu action='File'>"
       "<menuitem action='WindowNew'/>"
       "<menuitem action='TabNew'/>"
       "<separator/>"
       "<menuitem action='Open'/>"
       "<separator/>"
       "<menuitem action='SaveAs'/>"
       "<menuitem action='Send'/>"
       "<separator/>"
       "<menuitem action='TabClose'/>"
       "<menuitem action='WindowClose'/>"
       "<separator/>"
       "<menuitem action='PageSetup'/>"
       "<menuitem action='PrintPreview'/>"
       "<menuitem action='Print'/>"
       "<separator/>"
       "<menuitem action='PrivateBrowsing'/>"
       "<menuitem action='WorkOffline'/>"
       "<separator/>"
       "<menuitem action='Quit'/>"
      "</menu>"
      "<menu action='Edit'>"
       "<menuitem action='Undo'/>"
       "<menuitem action='Redo'/>"
       "<separator/>"
       "<menuitem action='Cut'/>"
       "<menuitem action='Copy'/>"
       "<menuitem action='Paste'/>"
       "<menuitem action='Delete'/>"
       "<menuitem action='SelectAll'/>"
       "<separator/>"
       "<menuitem action='FormFill'/>"
       "<separator/>"
       "<menuitem action='Find'/>"
       "<menuitem action='FindNext'/>"
       "<menuitem action='FindPrevious'/>"
       "<menuitem action='FindQuick'/>"
       "<separator/>"
       "<menuitem action='Preferences'/>"
      "</menu>"
      "<menu action='View'>"
       "<menuitem action='ToolbarNavigation'/>"
       "<menuitem action='ToolbarBookmarks'/>"
       // TODO: Does this belong here?
       "<menuitem action='Panels'/>"
       "<menuitem action='ToolbarDownloads'/>"
       "<menuitem action='ToolbarStatus'/>"
       "<separator/>"
       "<menuitem action='Refresh'/>"
       "<menuitem action='Stop'/>"
       "<menu action='RefreshEvery'>"
        "<menuitem action='RefreshEveryEnable'/>"
        "<separator/>"
        "<menuitem action='RefreshEvery30'/>"
        "<menuitem action='RefreshEvery60'/>"
        "<menuitem action='RefreshEvery300'/>"
        "<menuitem action='RefreshEvery900'/>"
        "<menuitem action='RefreshEvery1800'/>"
        "<menuitem action='RefreshEveryCustom'/>"
       "</menu>"
       "<separator/>"
       "<menuitem action='ZoomIn'/>"
       "<menuitem action='ZoomOut'/>"
       "<menuitem action='ZoomNormal'/>"
       "<separator/>"
       "<menuitem action='BackgroundImage'/>"
       "<menuitem action='SourceView'/>"
       "<menuitem action='Properties'/>"
       "<separator/>"
       "<menuitem action='SelectionSourceView'/>"
      "</menu>"
      "<menu action='Go'>"
       "<menuitem action='Back'/>"
       "<menuitem action='Forward'/>"
       "<menuitem action='Home'/>"
       "<menuitem action='Location'/>"
       "<menuitem action='Websearch'/>"
       // TODO: Is this a good idea?
       "<menuitem action='OpenInPanel'/>"
       "<menu action='TabsClosed'>"
        // Closed tabs shall be prepended here
        "<separator/>"
        "<menuitem action='TabsClosedClear'/>"
       "</menu>"
       "<menuitem action='UndoTabClose'/>"
       "<separator/>"
       "<menuitem action='SelectionLinksNewTabs'/>"
       "<menuitem action='SelectionTextTabNew'/>"
       "<menuitem action='SelectionTextTabCurrent'/>"
       "<menuitem action='SelectionTextWindowNew'/>"
       "<separator/>"
       "<menuitem action='SelectionSearch'/>"
       "<menuitem action='SelectionSearchWith'/>"
      "</menu>"
      "<menu action='Bookmarks'>"
       "<menuitem action='BookmarkNew'/>"
       "<menuitem action='BookmarksManage'/>"
       "<separator/>"
       // TODO: Insert bookmarks here
      "</menu>"
      "<menu action='Tools'>"
       "<menuitem action='PanelDownloads'/>"
       "<menuitem action='PanelBookmarks'/>"
       "<menuitem action='PanelHistory'/>"
       "<menuitem action='PanelTabs'/>"
       "<menuitem action='PanelCustom'/>"
       "<menuitem action='PanelExtensions'/>"
       "<menuitem action='PanelConsole'/>"
       "<separator/>"
       "<menuitem action='ColorPicker'/>"
       // TODO: Insert widgets and custom tools here
      "</menu>"
      "<menu action='Window'>"
       "<menuitem action='SessionLoad'/>"
       "<menuitem action='SessionSave'/>"
       "<separator/>"
       "<menuitem action='TabPrevious'/>"
       "<menuitem action='TabNext'/>"
       "<menuitem action='TabOverview'/>"
       "<separator/>"
       // All open tabs shall be appended here
      "</menu>"
      "<menu action='Help'>"
       "<menuitem action='HelpContents'/>"
       "<menuitem action='About'/>"
      "</menu>"
     "</menubar>"
     "<toolbar name='toolbar_navigation'>"
      "<toolitem action='TabNew'/>"
      "<toolitem action='Back'/>"
      "<toolitem action='Forward'/>"
      "<toolitem action='RefreshStop'/>"
      "<toolitem action='Home'/>"
      "<toolitem action='FormFill'/>"
      "<placeholder name='Location'/>"
      "<placeholder name='WebSearch'/>"
      "<placeholder name='TabTrash'/>"
     "</toolbar>"
     "<toolbar name='toolbar_panels'>"
      "<toolitem action='PanelDownloads'/>"
      "<toolitem action='PanelBookmarks'/>"
      "<toolitem action='PanelHistory'/>"
      "<toolitem action='PanelTabs'/>"
      "<toolitem action='PanelCustom'/>"
      "<toolitem action='PanelExtensions'/>"
      "<toolitem action='PanelConsole'/>"
     "</toolbar>"
     "<popup name='popup_tabsclosed'>"
      // Closed tabs shall be prepended here
      "<separator/>"
      "<menuitem action='TabsClosedClear'/>"
     "</popup>"
     "<popup name='popup_document'>"
      "<menuitem action='Menubar'/>"
      "<menuitem action='Back'/>"
      "<menuitem action='Forward'/>"
      "<menuitem action='Refresh'/>"
      "<menuitem action='Stop'/>"
      "<menu action='RefreshEvery'>"
       "<menuitem action='RefreshEveryEnable'/>"
       "<separator/>"
       "<menuitem action='RefreshEvery30'/>"
       "<menuitem action='RefreshEvery60'/>"
       "<menuitem action='RefreshEvery300'/>"
       "<menuitem action='RefreshEvery900'/>"
       "<menuitem action='RefreshEvery1800'/>"
       "<menuitem action='RefreshEveryCustom'/>"
      "</menu>"
      "<separator/>"
      "<menuitem action='LinkTabNew'/>"
      "<menuitem action='LinkTabCurrent'/>"
      "<menuitem action='LinkWindowNew'/>"
      "<separator/>"
      "<menuitem action='LinkBookmarkNew'/>"
      "<menuitem action='LinkSaveAs'/>"
      "<menuitem action='LinkSaveWith'/>"
      "<menuitem action='LinkCopy'/>"
      "<menuitem action='LinkSend'/>"
      "<separator/>"
      "<menuitem action='SelectionLinksNewTabs'/>"
      "<menuitem action='SelectionTextTabNew'/>"
      "<menuitem action='SelectionTextTabCurrent'/>"
      "<menuitem action='SelectionTextWindowNew'/>"
      "<separator/>"
      "<menuitem action='ImageViewTabNew'/>"
      "<menuitem action='ImageViewTabCurrent'/>"
      "<menuitem action='ImageSaveAs'/>"
      "<menuitem action='ImageSaveWith'/>"
      "<menuitem action='ImageCopy'/>"
      "<menuitem action='ImageSend'/>"
      "<separator/>"
      "<menuitem action='ImageViewTabNew'/>"
      "<menuitem action='ImageViewTabCurrent'/>"
      "<separator/>"
      "<menuitem action='UndoTabClose'/>"
      "<menuitem action='BookmarkNew'/>"
      "<menuitem action='SaveAs'/>"
      "<menuitem action='Send'/>"
      "<menuitem action='Print'/>"
      "<separator/>"
      "<menuitem action='Copy'/>"
      "<menuitem action='SelectAll'/>"
      "<menuitem action='SelectionSearch'/>"
      "<menuitem action='SelectionSearchWith'/>"
      "<menuitem action='SelectionSourceView'/>"
      "<separator/>"
      "<menuitem action='BackgroundImage'/>"
      "<menuitem action='SourceView'/>"
      "<menuitem action='Properties'/>"
     "</popup>"
    "</ui>";

    // Let us see some ui manager magic
    browser->actiongroup = gtk_action_group_new("Browser");
    gtk_action_group_add_actions(browser->actiongroup, entries, entries_n, browser);
    gtk_action_group_add_toggle_actions(browser->actiongroup
     , toggle_entries, toggle_entries_n, browser);
    gtk_action_group_add_radio_actions(browser->actiongroup
     , refreshevery_entries, refreshevery_entries_n
     , 300, NULL/*G_CALLBACK(activate_refreshevery_period_action)*/, browser);
    GtkUIManager* ui_manager = gtk_ui_manager_new();
    gtk_ui_manager_insert_action_group(ui_manager, browser->actiongroup, 0);
    gtk_window_add_accel_group(GTK_WINDOW(browser->window)
     , gtk_ui_manager_get_accel_group(ui_manager));

    GError* error = NULL;
    if(!gtk_ui_manager_add_ui_from_string(ui_manager, ui_markup, -1, &error))
    {
     // TODO: Should this be a message dialog? When does this happen?
     g_message("User interface couldn't be created: %s", error->message);
     g_error_free(error);
    }

    // Make all actions except toplevel menus which lack a callback insensitive
    // This will vanish once all actions are implemented
    gint i; for(i = 0; i < entries_n; i++)
    {
     GtkAction* action = gtk_action_group_get_action(browser->actiongroup, entries[i].name);
     gtk_action_set_sensitive(action, entries[i].callback || !entries[i].tooltip);
    }
    for(i = 0; i < toggle_entries_n; i++)
    {
     GtkAction* action = gtk_action_group_get_action(browser->actiongroup
      , toggle_entries[i].name);
     gtk_action_set_sensitive(action, toggle_entries[i].callback != NULL);
    }
    for(i = 0; i < refreshevery_entries_n; i++)
    {
     GtkAction* action = gtk_action_group_get_action(browser->actiongroup
      , refreshevery_entries[i].name);
     gtk_action_set_sensitive(action, FALSE);
    }

    action_set_sensitive("PrivateBrowsing", FALSE, browser);
    action_set_sensitive("WorkOffline", FALSE, browser);

    //action_set_active("Menubar", config->bMenubar, browser);
    action_set_active("ToolbarNavigation", config->bToolbarNavigation, browser);
    action_set_active("ToolbarBookmarks", config->bToolbarBookmarks, browser);
    //action_set_active("ToolbarDownloads", config->bToolbarDownloads, browser);
    action_set_active("ToolbarStatus", config->bToolbarStatus, browser);

    // Create the menubar
    browser->menubar = gtk_ui_manager_get_widget(ui_manager, "/menubar");
    gtk_box_pack_start(GTK_BOX(vbox), browser->menubar, FALSE, FALSE, 0);
    browser->menu_tabsclosed = gtk_ui_manager_get_widget(ui_manager, "/popup_tabsclosed");
    GtkWidget* menuitem = gtk_ui_manager_get_widget(ui_manager, "/menubar/Go/TabsClosed");
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem), browser->menu_tabsclosed);
    menuitem = gtk_separator_menu_item_new();
    gtk_widget_show(menuitem);
    gtk_menu_shell_prepend(GTK_MENU_SHELL(browser->menu_tabsclosed), menuitem);
    browser->menu_window = gtk_menu_item_get_submenu(
     GTK_MENU_ITEM(gtk_ui_manager_get_widget(ui_manager, "/menubar/Window")));
    menuitem = gtk_separator_menu_item_new();
    gtk_widget_show(menuitem);
    gtk_menu_shell_append(GTK_MENU_SHELL(browser->menu_window), menuitem);
    browser->menu_document = gtk_ui_manager_get_widget(ui_manager, "/popup_document");
    g_object_ref(browser->menu_document);

    // Create the navigation toolbar
    browser->navibar = gtk_ui_manager_get_widget(ui_manager, "/toolbar_navigation");
    gtk_toolbar_set_style(GTK_TOOLBAR(browser->navibar)
     , config_to_toolbarstyle(config->iToolbarStyle));
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(browser->navibar)
     , config_to_toolbariconsize(config->iToolbarIconSize));
    gtk_toolbar_set_show_arrow(GTK_TOOLBAR(browser->navibar), TRUE);
    gtk_box_pack_start(GTK_BOX(vbox), browser->navibar, FALSE, FALSE, 0);
    GtkAction* action = gtk_action_group_get_action(browser->actiongroup, "Back");
    g_object_set(action, "is-important", TRUE, NULL);

    // Location entry
    browser->navibar_location = sexy_icon_entry_new();
    entry_setup_completion(GTK_ENTRY(browser->navibar_location));
    sokoke_entry_set_can_undo(GTK_ENTRY(browser->navibar_location), TRUE);
    browser->icon_page = gtk_image_new();
    sexy_icon_entry_set_icon(SEXY_ICON_ENTRY(browser->navibar_location)
     , SEXY_ICON_ENTRY_PRIMARY, GTK_IMAGE(browser->icon_page));
    browser->icon_lock = gtk_image_new();
    sexy_icon_entry_set_icon(SEXY_ICON_ENTRY(browser->navibar_location)
     , SEXY_ICON_ENTRY_SECONDARY, GTK_IMAGE(browser->icon_lock));
    g_signal_connect(browser->navibar_location, "key-press-event"
     , G_CALLBACK(on_navibar_entry_location_key_down), browser);
    GtkToolItem* toolitem = gtk_tool_item_new();
    gtk_tool_item_set_expand(GTK_TOOL_ITEM(toolitem), TRUE);
    gtk_container_add(GTK_CONTAINER(toolitem), browser->navibar_location);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->navibar), toolitem, -1);

    // Search entry
    // TODO: Paint an arrow over the icon
    // Make this actively resizable or enlarge to fit contents?
    // TODO: custom icon
    // select search engine via popup or keyboard or mouse wheel
    // checkbutton 'Open in new tab' in the popup
    browser->navibar_search = sexy_icon_entry_new();
    // TODO: Display "show in context menu" search engines as "completion actions"
    entry_setup_completion(GTK_ENTRY(browser->navibar_search));
    sokoke_entry_set_can_undo(GTK_ENTRY(browser->navibar_search), TRUE);
    sokoke_entry_set_default_text(GTK_ENTRY(browser->navibar_search)
     , "Search the web");
    browser->icon_search = gtk_image_new_from_stock(GTK_STOCK_FIND
     , GTK_ICON_SIZE_MENU);
    sexy_icon_entry_set_icon(SEXY_ICON_ENTRY(browser->navibar_search)
     , SEXY_ICON_ENTRY_PRIMARY, GTK_IMAGE(browser->icon_search));
    browser->icon_search_add = gtk_image_new();
    sexy_icon_entry_set_icon(SEXY_ICON_ENTRY(browser->navibar_search)
     , SEXY_ICON_ENTRY_SECONDARY, GTK_IMAGE(browser->icon_search_add));
    g_signal_connect(browser->navibar_search, "activate"
     , G_CALLBACK(on_navibar_entry_search_activate), browser);
    toolitem = gtk_tool_item_new();
    gtk_container_add(GTK_CONTAINER(toolitem), browser->navibar_search);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->navibar), toolitem, -1);
    action = gtk_action_group_get_action(browser->actiongroup, "TabsClosed");
    toolitem = GTK_TOOL_ITEM(gtk_action_create_tool_item(action));
    g_signal_connect(toolitem, "clicked", G_CALLBACK(on_navibar_tabsclosed_clicked), browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->navibar), toolitem, -1);
    sokoke_container_show_children(GTK_CONTAINER(browser->navibar));

    // Bookmarkbar
    browser->bookmarkbar = gtk_toolbar_new();
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(browser->bookmarkbar), GTK_ICON_SIZE_MENU);
    gtk_toolbar_set_style(GTK_TOOLBAR(browser->bookmarkbar), GTK_TOOLBAR_BOTH_HORIZ);
    toolitem = tool_button_new("Folder", GTK_STOCK_DIRECTORY, TRUE, TRUE
     , NULL, NULL, browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->bookmarkbar), toolitem, -1);
    toolitem = tool_button_new("Example", STOCK_BOOKMARK, TRUE, TRUE
     , NULL/*on_menu_bookmarks_item_activate*/, "http://example.com", browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->bookmarkbar), toolitem, -1);
    toolitem = tool_button_new("Example", STOCK_BOOKMARK, TRUE, TRUE
     , NULL/*on_menu_bookmarks_item_activate*/, "http://example.com", browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->bookmarkbar), toolitem, -1);
    sokoke_container_show_children(GTK_CONTAINER(browser->bookmarkbar));
    gtk_box_pack_start(GTK_BOX(vbox), browser->bookmarkbar, FALSE, FALSE, 0);

    // Superuser warning
    #ifdef G_OS_UNIX
    if(G_UNLIKELY(!geteuid())) // effective superuser?
    {
     hbox = gtk_event_box_new();
     gtk_widget_modify_bg(hbox, GTK_STATE_NORMAL
      , &hbox->style->bg[GTK_STATE_SELECTED]);
     label = gtk_label_new("Warning: You are using the superuser account!");
     gtk_misc_set_padding(GTK_MISC(label), 0, 2);
     gtk_widget_modify_fg(GTK_WIDGET(label), GTK_STATE_NORMAL
      , &GTK_WIDGET(label)->style->fg[GTK_STATE_SELECTED]);
     gtk_container_add(GTK_CONTAINER(hbox), GTK_WIDGET(label));
     gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    }
    #endif

    // Create the panels
    GtkWidget* hpaned = gtk_hpaned_new();
    gtk_paned_set_position(GTK_PANED(hpaned), config->iPanelPos);
    g_signal_connect(hpaned, "notify::position"
     , GTK_SIGNAL_FUNC(on_panels_notify_position_activate), browser);
    gtk_box_pack_start(GTK_BOX(vbox), hpaned, TRUE, TRUE, 0);
    gtk_widget_show(hpaned);

    browser->panels = gtk_hbox_new(FALSE, 0);
    gtk_paned_pack1(GTK_PANED(hpaned), browser->panels, FALSE, FALSE/*TRUE*/);
    sokoke_widget_set_visible(browser->panels, config->bPanels);

    // Create the panel toolbar
    GtkWidget* panelbar = gtk_ui_manager_get_widget(ui_manager, "/toolbar_panels");
    gtk_toolbar_set_style(GTK_TOOLBAR(panelbar), GTK_TOOLBAR_BOTH);
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(panelbar), GTK_ICON_SIZE_BUTTON);
    gtk_toolbar_set_orientation(GTK_TOOLBAR(panelbar), GTK_ORIENTATION_VERTICAL); 
    gtk_box_pack_start(GTK_BOX(browser->panels), sokoke_frame_new_child_shadow(
     panelbar, GTK_SHADOW_OUT), FALSE, FALSE, 0);
    gtk_widget_show(gtk_widget_get_parent(panelbar));
    g_object_unref(ui_manager);

    GtkWidget* cbox = gtk_vbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(browser->panels), cbox, TRUE, TRUE, 0);
    gtk_widget_show(cbox);

    // Panels title bar
    GtkWidget* labelbar = gtk_toolbar_new();
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(labelbar), GTK_ICON_SIZE_MENU);
    gtk_toolbar_set_style(GTK_TOOLBAR(labelbar), GTK_TOOLBAR_ICONS);
    toolitem = gtk_tool_item_new();
    gtk_tool_item_set_expand(toolitem, TRUE);
    label = gtk_label_new_with_mnemonic("_Panels");
    gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
    gtk_container_add(GTK_CONTAINER(toolitem), label);
    gtk_container_set_border_width(GTK_CONTAINER(toolitem), 6);
    gtk_toolbar_insert(GTK_TOOLBAR(labelbar), toolitem, -1);
    // TODO: Does 'goto top' actually indicate 'detach'?
    toolitem = tool_button_new(NULL, GTK_STOCK_GOTO_TOP, FALSE, TRUE
     , NULL/*G_CALLBACK(on_panels_button_float_clicked)*/, "Detach panel", browser);
    gtk_toolbar_insert(GTK_TOOLBAR(labelbar), toolitem, -1);
    toolitem = tool_button_new(NULL, GTK_STOCK_CLOSE, FALSE, TRUE
     , G_CALLBACK(on_panels_button_close_clicked), "Close panel", browser);
    gtk_toolbar_insert(GTK_TOOLBAR(labelbar), toolitem, -1);
    gtk_box_pack_start(GTK_BOX(cbox), labelbar, FALSE, FALSE, 0);
    gtk_widget_show_all(labelbar);

    // Notebook, containing all panels
    browser->panels_notebook = gtk_notebook_new();
    gtk_notebook_set_show_border(GTK_NOTEBOOK(browser->panels_notebook), FALSE);
    gtk_notebook_set_show_tabs(GTK_NOTEBOOK(browser->panels_notebook), FALSE);
      // Custom
      #ifdef MIDORI_WEBKIT
      GtkWidget* scrolled = gtk_scrolled_window_new(NULL, NULL);
      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled)
       , GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
      GTK_WIDGET_SET_FLAGS(scrolled, GTK_CAN_FOCUS);
      browser->panel_custom = webkit_gtk_page_new();
      #else
      GtkWidget* scrolled = gtk_event_box_new();
      browser->panel_custom = webi_new();
      #endif
      gtk_container_add(GTK_CONTAINER(scrolled), browser->panel_custom);
      document_load_uri(browser->panel_custom, config->sPanelCustom);
      gtk_notebook_append_page(GTK_NOTEBOOK(browser->panels_notebook)
       , scrolled, NULL);
    gtk_box_pack_start(GTK_BOX(cbox), browser->panels_notebook, TRUE, TRUE, 0);
    gtk_widget_show_all(browser->panels_notebook);
    action_set_active("PanelCustom", config->bToolbarStatus, browser);//...

    // Notebook, containing all documents
    browser->documents = gtk_notebook_new();
    gtk_notebook_set_scrollable(GTK_NOTEBOOK(browser->documents), TRUE);
    #if GTK_CHECK_VERSION(2, 10, 0)
    //gtk_notebook_set_group_id(GTK_NOTEBOOK(browser->documents), 0);
    #endif
    g_signal_connect_after(GTK_OBJECT(browser->documents), "switch-page"
     , GTK_SIGNAL_FUNC(on_notebook_switch_page), browser);
    gtk_paned_pack2(GTK_PANED(hpaned), browser->documents, FALSE, FALSE);
    gtk_widget_show(browser->documents);

    // Incremental findbar
    browser->findbox = gtk_toolbar_new();
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(browser->findbox), GTK_ICON_SIZE_MENU);
    gtk_toolbar_set_style(GTK_TOOLBAR(browser->findbox), GTK_TOOLBAR_BOTH_HORIZ);
    toolitem = gtk_tool_item_new();
    gtk_container_set_border_width(GTK_CONTAINER(toolitem), 6);
    gtk_container_add(GTK_CONTAINER(toolitem)
     , gtk_label_new_with_mnemonic("_Inline find:"));
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    browser->findbox_text = sexy_icon_entry_new();
    // TODO: Change that to an error icon if a search fails
    GtkWidget* icon = gtk_image_new_from_stock(GTK_STOCK_FIND, GTK_ICON_SIZE_MENU);
    sexy_icon_entry_set_icon(SEXY_ICON_ENTRY(browser->findbox_text)
     , SEXY_ICON_ENTRY_PRIMARY, GTK_IMAGE(icon));
    sexy_icon_entry_add_clear_button(SEXY_ICON_ENTRY(browser->findbox_text));
    sokoke_entry_set_can_undo(GTK_ENTRY(browser->findbox_text), TRUE);
    g_signal_connect(browser->findbox_text, "activate"
     , GTK_SIGNAL_FUNC(on_action_find_next_activate), browser);
    toolitem = gtk_tool_item_new();
    gtk_container_add(GTK_CONTAINER(toolitem), browser->findbox_text);
    gtk_tool_item_set_expand(GTK_TOOL_ITEM(toolitem), TRUE);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    toolitem = tool_button_new(NULL, GTK_STOCK_GO_BACK, TRUE, TRUE
     , G_CALLBACK(on_action_find_previous_activate), NULL, browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    toolitem = tool_button_new(NULL, GTK_STOCK_GO_FORWARD, TRUE, TRUE
     , G_CALLBACK(on_action_find_next_activate), NULL, browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    browser->findbox_case = gtk_toggle_tool_button_new_from_stock(GTK_STOCK_SPELL_CHECK);
    gtk_tool_button_set_label(GTK_TOOL_BUTTON(browser->findbox_case), "Match case");
    gtk_tool_item_set_is_important(GTK_TOOL_ITEM(browser->findbox_case), TRUE);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), browser->findbox_case, -1);
    toolitem = gtk_separator_tool_item_new();
    gtk_separator_tool_item_set_draw(GTK_SEPARATOR_TOOL_ITEM(toolitem), FALSE);
    gtk_tool_item_set_expand(GTK_TOOL_ITEM(toolitem), TRUE);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    toolitem = tool_button_new(NULL, GTK_STOCK_CLOSE, FALSE, TRUE
     , G_CALLBACK(on_findbox_button_close_clicked), "Close findbar", browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    sokoke_container_show_children(GTK_CONTAINER(browser->findbox));
    gtk_box_pack_start(GTK_BOX(vbox), browser->findbox, FALSE, FALSE, 0);

    // TODO: fix children overlapping statusbar border
    browser->statusbar = gtk_statusbar_new();
    gtk_box_pack_start(GTK_BOX(vbox), browser->statusbar, FALSE, FALSE, 0);
    browser->progress = gtk_progress_bar_new();
    // Setting the progressbar's height to 1 makes it fit in the statusbar
    gtk_widget_set_size_request(browser->progress, -1, 1);
    gtk_box_pack_start(GTK_BOX(browser->statusbar), browser->progress
     , FALSE, FALSE, 3);

    browser->icon_link_script = gtk_image_new_from_icon_name(
     STOCK_SCRIPT, GTK_ICON_SIZE_MENU);
    gtk_box_pack_start(GTK_BOX(browser->statusbar), browser->icon_link_script
     , FALSE, FALSE, 0);
    browser->icon_link_window = gtk_image_new_from_icon_name(
     STOCK_WINDOW_NEW, GTK_ICON_SIZE_MENU);
    gtk_box_pack_start(GTK_BOX(browser->statusbar), browser->icon_link_window
     , FALSE, FALSE, 0);
    // TODO: Show network status here if possible
    browser->icon_network = gtk_image_new_from_icon_name(
     STOCK_NETWORK_OFFLINE, GTK_ICON_SIZE_MENU);
    //gtk_widget_show(browser->icon_network);
    gtk_box_pack_start(GTK_BOX(browser->statusbar), browser->icon_network
     , FALSE, FALSE, 0);

    browser->misc_feed = gtk_hbox_new(FALSE, 0);
    icon = gtk_image_new_from_icon_name(GTK_STOCK_INDEX, GTK_ICON_SIZE_MENU);
    gtk_box_pack_start(GTK_BOX(browser->misc_feed), icon, FALSE, FALSE, 0);
    label = gtk_label_new("News feed available");
    gtk_box_pack_start(GTK_BOX(browser->misc_feed), label, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(browser->statusbar), browser->misc_feed
     , FALSE, FALSE, 0);

    }
    else
    {

    browser->window = oldBrowser->window;
    browser->menubar = oldBrowser->menubar;
    browser->actiongroup = oldBrowser->actiongroup;
    browser->menu_tabsclosed = oldBrowser->menu_tabsclosed;
    browser->menu_window = oldBrowser->menu_window;
    browser->menu_document = oldBrowser->menu_document;
    browser->navibar = oldBrowser->navibar;
    browser->icon_page = oldBrowser->icon_page;
    browser->navibar_location = oldBrowser->navibar_location;
    browser->icon_lock = oldBrowser->icon_lock;
    browser->icon_search = oldBrowser->icon_search;
    browser->icon_search_add = oldBrowser->icon_search_add;
    browser->navibar_search = oldBrowser->navibar_search;
    browser->bookmarkbar = oldBrowser->bookmarkbar;
    browser->panels = oldBrowser->panels;
    browser->panels_notebook = oldBrowser->panels_notebook;
    browser->panel_custom = oldBrowser->panel_custom;
    browser->documents = oldBrowser->documents;
    browser->findbox = oldBrowser->findbox;
    browser->statusbar = oldBrowser->statusbar;
    browser->icon_link_script = oldBrowser->icon_link_script;
    browser->icon_link_window = oldBrowser->icon_link_window;
    browser->icon_network = oldBrowser->icon_network;
    browser->progress = oldBrowser->progress;
    browser->misc_feed = oldBrowser->misc_feed;

    }

    // Define some default values
    browser->bMenubar = TRUE;
    browser->bToolbar = TRUE;
    browser->bLocation = TRUE;
    browser->bStatusbar = TRUE;
    browser->bResizable = TRUE;
    browser->sUri = g_strdup("about:blank");
    browser->sTitle = g_strdup("about:blank");
    browser->iLoadedPercent = -1; // initially "not loading"

    // Add a window menu item
    // TODO: Menu items should be ordered like the notebook tabs
    // TODO: Watch tab reordering in >= gtk 2.10
    browser->document_menu = menu_item_new(browser->sTitle, GTK_STOCK_FILE, "", NULL
     , (GdkModifierType)0, G_CALLBACK(on_window_menu_item_activate), TRUE, browser);
    gtk_widget_show(browser->document_menu);
    gtk_menu_shell_append(GTK_MENU_SHELL(browser->menu_window), browser->document_menu);

    // Create a new page
    GtkWidget* eventbox = gtk_event_box_new();
    gtk_event_box_set_visible_window(GTK_EVENT_BOX(eventbox), FALSE);
    g_signal_connect(GTK_OBJECT(eventbox), "button_press_event"
     , GTK_SIGNAL_FUNC(on_notebook_tab_mouse_down), browser);
    GtkWidget* hbox = gtk_hbox_new(FALSE, 1);
    gtk_container_add(GTK_CONTAINER(eventbox), GTK_WIDGET(hbox));
    browser->document_icon = gtk_image_new_from_stock(GTK_STOCK_FILE, GTK_ICON_SIZE_MENU);
    gtk_box_pack_start(GTK_BOX(hbox), browser->document_icon, FALSE, FALSE, 0);
    browser->document_name = gtk_label_new(browser->sTitle);
    gtk_misc_set_alignment(GTK_MISC(browser->document_name), 0.0, 0.5);
    // TODO: make the tab initially look "unvisited" until it's focused
    // TODO: tabs should shrink when there is not enough space
    // TODO: gtk's tab scrolling is weird?
    gint w, h; sokoke_widget_get_text_size(browser->document_name, "M", &w, &h);
    gtk_widget_set_size_request(GTK_WIDGET(browser->document_name)
     , w * config->iTabSize, -1);
    gtk_label_set_ellipsize(GTK_LABEL(browser->document_name), PANGO_ELLIPSIZE_END);
    gtk_box_pack_start(GTK_BOX(hbox), browser->document_name, FALSE, FALSE, 0);
    browser->document_close = gtk_button_new();
    gtk_button_set_relief(GTK_BUTTON(browser->document_close), GTK_RELIEF_NONE);
    gtk_button_set_focus_on_click(GTK_BUTTON(browser->document_close), FALSE);
    GtkRcStyle* rcstyle = gtk_rc_style_new();
    rcstyle->xthickness = rcstyle->ythickness = 0;
    gtk_widget_modify_style(browser->document_close, rcstyle);
    gtk_rc_style_unref(rcstyle);
    GtkWidget* image = gtk_image_new_from_stock(GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU);
    gtk_button_set_image(GTK_BUTTON(browser->document_close), image);
    gtk_box_pack_start(GTK_BOX(hbox), browser->document_close, FALSE, FALSE, 0);
    GtkSettings* gtksettings = gtk_settings_get_default(); gint height;
    gtk_icon_size_lookup_for_settings(gtksettings, GTK_ICON_SIZE_BUTTON, 0, &height);
    gtk_widget_set_size_request(browser->document_close, -1, height);
    gtk_widget_show_all(GTK_WIDGET(eventbox));
    sokoke_widget_set_visible(browser->document_close, config->iTabClose);
    g_signal_connect(GTK_OBJECT(browser->document_close), "button_press_event"
     , GTK_SIGNAL_FUNC(on_notebook_tab_close_mouse_down), browser);
    #ifdef MIDORI_WEBKIT
    GtkWidget* scrolled = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled)
     , GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    GTK_WIDGET_SET_FLAGS(scrolled, GTK_CAN_FOCUS);
    #else
    GtkWidget* scrolled = gtk_event_box_new();
    #endif
    g_signal_connect(GTK_OBJECT(scrolled), "button_press_event"
     , GTK_SIGNAL_FUNC(on_document_mouse_down), browser);
    g_signal_connect(GTK_OBJECT(scrolled), "popup-menu"
     , GTK_SIGNAL_FUNC(on_document_popup), browser);
    g_signal_connect(GTK_OBJECT(scrolled), "scroll-event"
     , GTK_SIGNAL_FUNC(on_document_scroll), browser);
    gint page = gtk_notebook_insert_page(GTK_NOTEBOOK(browser->documents)
     , scrolled, GTK_WIDGET(eventbox)
     , gtk_notebook_get_current_page(GTK_NOTEBOOK(browser->documents)) + 1);
    #if GTK_CHECK_VERSION(2, 10, 0)
    gtk_notebook_set_tab_reorderable(GTK_NOTEBOOK(browser->documents)
     , scrolled, TRUE);
    gtk_notebook_set_tab_detachable(GTK_NOTEBOOK(browser->documents)
     , scrolled, TRUE);
    #endif

    // Create a new document
    #define DOC_CONNECT(__sig, __func) g_signal_connect(\
     G_OBJECT(browser->document), __sig\
     , G_CALLBACK(__func), browser);
    gchar* sAgentName = g_strconcat(PACKAGE_NAME, "/", PACKAGE_VERSION, NULL);
    #ifdef MIDORI_WEBI
    browser->document = webi_new();
    webi_set_emit_internal_status(WEBI(browser->document), TRUE); // crap
    gtk_widget_show_all(browser->document); // bug
    DOC_CONNECT("location", on_document_location_changed)
    DOC_CONNECT("title", on_document_title_changed)
    DOC_CONNECT("load-start", on_document_load_started)
    DOC_CONNECT("load-stop", on_document_load_finished)
    DOC_CONNECT("status", on_document_load_changed)
    DOC_CONNECT("selection", on_document_selection_changed)
    DOC_CONNECT("req-js-prompt", on_document_request_script_prompt)
    DOC_CONNECT("req-auth-prompt", on_document_request_auth_prompt)
    DOC_CONNECT("req-new-window", on_document_window_open)
    DOC_CONNECT("set-window-properties", on_document_window_features)
    DOC_CONNECT("set-cookie", on_document_set_cookie)
    DOC_CONNECT("mouse-over", on_document_link_hover)
    #elif defined(MIDORI_WEBKIT)
    browser->document = GTK_WIDGET(webkit_gtk_page_new());
    WebKitGtkFrame* mainframe = webkit_gtk_page_get_main_frame(
     WEBKIT_GTK_PAGE(browser->document));
    #define FRAME_CONNECT(__sig, __func) g_signal_connect(\
     G_OBJECT(mainframe), __sig\
     , G_CALLBACK(__func), browser);
    /*webkit_gtk_page_set_application_name_for_user_agent(
     PACKAGE_NAME, PACKAGE_VERSION);*/
    //webkit_gtk_page_set_custom_user_agent(sAgentName);
    FRAME_CONNECT("title-changed", on_frame_title_changed)
    DOC_CONNECT("title-changed", on_document_title_changed)
    DOC_CONNECT("icon-loaded", on_document_icon_changed)
    DOC_CONNECT("load-started", on_document_load_started)
    DOC_CONNECT("load-progress-changed", on_document_load_changed)
    DOC_CONNECT("load-finished", on_document_load_finished)
    FRAME_CONNECT("cleared", on_frame_cleared)
    FRAME_CONNECT("load-done", on_frame_load_done)
    DOC_CONNECT("status-bar-text-changed", on_document_status_message)
    #endif
    g_free(sAgentName);

    DOC_CONNECT("destroy", on_document_destroy)
    #undef DOC_CONNECT

    // Eventually pack and display everything
    gtk_container_add(GTK_CONTAINER(scrolled), browser->document);
    sokoke_widget_set_visible(browser->menubar, TRUE/*config->bShowMenubar*/);
    sokoke_widget_set_visible(browser->navibar, config->bToolbarNavigation);
    sokoke_widget_set_visible(browser->bookmarkbar, config->bToolbarBookmarks);
    sokoke_widget_set_visible(browser->statusbar, config->bToolbarStatus);
    gtk_widget_show(GTK_WIDGET(scrolled));
    gtk_widget_show(GTK_WIDGET(browser->document));
    gtk_notebook_set_current_page(GTK_NOTEBOOK(browser->documents), page);

    update_browser_actions(browser);
    gtk_widget_show(browser->window);
    gtk_widget_grab_focus(GTK_WIDGET(browser->navibar_location));

    return browser;
}

// -- configuration related

static CConfig* config_load(const gchar* sFilename, GError* *error)
{
    DEBUG_FUNC
    CConfig* config = g_new0(CConfig, 1);

    GKeyFile* keyFile = g_key_file_new();
    g_key_file_load_from_file(keyFile, sFilename
     , G_KEY_FILE_KEEP_COMMENTS, error);
    #define GET_INT(var, key, default) \
     var = sokoke_key_file_get_integer_default( \
     keyFile, "browser", key, default, NULL)
    #define GET_STR(var, key, default) \
     var = sokoke_key_file_get_string_default( \
     keyFile, "browser", key, default, NULL)
    GET_STR(config->sHomepage, "Homepage", "http://www.google.com");
    GET_STR(config->sLocationSearch, "LocationSearch", "http://www.google.com/search?q=%s");
    GET_INT(config->bToolbarNavigation, "ToolbarNavigation", TRUE);
    GET_INT(config->bToolbarBookmarks, "ToolbarBookmarks", FALSE);
    GET_INT(config->bToolbarStatus, "ToolbarStatus", TRUE);
    //GET_INT(config->bToolbarDownloads, "ToolbarDownloads", FALSE);
    GET_INT(config->iToolbarStyle, "ToolbarStyle", 0);
    GET_INT(config->iToolbarIconSize, "ToolbarIconSize", 0);
    GET_INT(config->bPanels, "Panels", FALSE);
    GET_INT(config->iPanelPos, "PanelPos", 0);
    GET_STR(config->sPanelCustom, "PanelCustom", "http://www.google.com");
    GET_INT(config->iTabSize, "TabSize", 10);
    GET_INT(config->iTabClose, "TabClose", 1);
    GET_INT(config->iPopupRule, "PopupRule", 2);
    #undef GET_INT
    #undef GET_STR

    return config;
}

gboolean config_save(CConfig* config, const gchar* sFilename, GError* *error)
{
    DEBUG_FUNC
    GKeyFile* keyFile = g_key_file_new();

    g_key_file_set_string(keyFile, "browser", "Homepage", config->sHomepage);
    g_key_file_set_string(keyFile, "browser", "LocationSearch", config->sLocationSearch);
    g_key_file_set_integer(keyFile, "browser", "ToolbarNavigation", config->bToolbarNavigation);
    g_key_file_set_integer(keyFile, "browser", "ToolbarBookmarks", config->bToolbarBookmarks);
    //g_key_file_set_integer(keyFile, "browser", "ToolbarDownloads", config->bToolbarDownloads);
    g_key_file_set_integer(keyFile, "browser", "ToolbarStatus", config->bToolbarStatus);
    g_key_file_set_integer(keyFile, "browser", "ToolbarStyle", config->iToolbarStyle);
    g_key_file_set_integer(keyFile, "browser", "ToolbarIconSize", config->iToolbarIconSize);
    g_key_file_set_integer(keyFile, "browser", "Panels", config->bPanels);
    g_key_file_set_integer(keyFile, "browser", "PanelPos", config->iPanelPos);
    g_key_file_set_string(keyFile, "browser", "PanelCustom", config->sPanelCustom);
    g_key_file_set_integer(keyFile, "browser", "TabSize", config->iTabSize);
    g_key_file_set_integer(keyFile, "browser", "TabClose", config->iTabClose);
    g_key_file_set_integer(keyFile, "browser", "PopupRule", config->iPopupRule);

    //g_key_file_set_string(keyFile, "content", "LoadImages", config->bLoadImages);
    //g_key_file_set_string(keyFile, "content", "PrintBackgrounds", config->bPrintBackgrounds);
    //g_key_file_set_string(keyFile, "content", "JavaScript", config->bJavaScript);
    //g_key_file_set_string(keyFile, "content", "StandardFont", config->sStandardFont);
    //g_key_file_set_string(keyFile, "content", "SerifFont", config->sSerifFont);
    //g_key_file_set_string(keyFile, "content", "SansFont", config->sSansFontFont);
    //g_key_file_set_string(keyFile, "content", "FixedFont", config->sFixedFont);
    //g_key_file_set_string(keyFile, "content", "MinFontSize", config->iMinFontSize);

    gboolean bSaved = sokoke_key_file_save_to_file(keyFile, sFilename, error);
    g_key_file_free(keyFile);

    return bSaved;
}

// -- stock icons

void init_stock_icons(void)
{
    // Based on ephy_stock_icons_init from lib/ephy-stock-icons.c from epiphany.
    // Copyright (C) 2002 Jorn Baayen
    GtkIconFactory *factory;
    GtkIconSet *icon_set;
    GtkIconSource *icon_source;
    int i;
    static const gchar* icon_theme_items[] =
    {
       STOCK_LOCK_OPEN
     , STOCK_LOCK_SECURE
     , STOCK_LOCK_BROKEN
     , STOCK_SCRIPT
     , STOCK_THEME
     , STOCK_USER_TRASH
    };
    static const GtkStockItem items[] =
    {
       { STOCK_BOOKMARK,         "Bookmark", 0, 0, NULL }
     , { STOCK_BOOKMARK_NEW,     "New bookmark", 0, 0, NULL }
     , { STOCK_BOOKMARKS,        "_Bookmarks", 0, 0, NULL }
     , { STOCK_DOWNLOADS,        "_Downloads", 0, 0, NULL }
     , { STOCK_CONSOLE,          "_Console", 0, 0, NULL }
     , { STOCK_EXTENSIONS,       "_Extensions", 0, 0, NULL }
     , { STOCK_FORM_FILL,        "_Form fill", 0, 0, NULL }
     , { STOCK_HISTORY,          "History", 0, 0, NULL }
     , { STOCK_HOMEPAGE,         "Homepage", 0, 0, NULL }
     , { STOCK_LOCATION,         "Location entry", 0, 0, NULL }
     , { STOCK_NEWSFEED,         "Newsfeed", 0, 0, NULL }
     , { STOCK_PLUGINS,          "_Plugins", 0, 0, NULL }
     , { STOCK_POPUPS_BLOCKED,   "Blocked popups", 0, 0, NULL }
     , { STOCK_SOURCE_VIEW,      "View source", 0, 0, NULL }
     , { STOCK_TAB_CLOSE,        "Close tab", 0, 0, NULL }
     , { STOCK_TAB_NEW,          "New tab", 0, 0, NULL }
     , { STOCK_WINDOW_CLOSE,     "Close window", 0, 0, NULL }
     , { STOCK_WINDOW_NEW,       "New window", 0, 0, NULL }
     #if !GTK_CHECK_VERSION(2, 10, 0)
     , { GTK_STOCK_SELECT_ALL,   "_Select all", 0, 0, NULL }
     #endif
    };
    // TODO: Use fallback icons with _ suffix
    factory = gtk_icon_factory_new ();
    for(i = 0; i < (int)G_N_ELEMENTS(items); i++)
    {
     icon_source = gtk_icon_source_new();
     gtk_icon_source_set_icon_name (icon_source, items[i].stock_id);
     icon_set = gtk_icon_set_new();
     gtk_icon_set_add_source(icon_set, icon_source);
     gtk_icon_source_free(icon_source);
     gtk_icon_factory_add(factory, items[i].stock_id, icon_set);
     gtk_icon_set_unref(icon_set);
    }
    gtk_stock_add_static(items, G_N_ELEMENTS(items));

    for(i = 0; i < (int)G_N_ELEMENTS(icon_theme_items); i++)
    {
     icon_source = gtk_icon_source_new();
     gtk_icon_source_set_icon_name(icon_source, icon_theme_items[i]);
     icon_set = gtk_icon_set_new();
     gtk_icon_set_add_source(icon_set, icon_source);
     gtk_icon_source_free(icon_source);
     gtk_icon_factory_add(factory, icon_theme_items[i], icon_set);
     gtk_icon_set_unref(icon_set);
    }
    gtk_icon_factory_add_default(factory);
    g_object_unref(factory);

    // TODO: multi-head?
    // TODO: Do we need that stuff?
    /*gtk_icon_theme_append_search_path(gtk_icon_theme_get_default()
     , SHARE_DIR G_DIR_SEPARATOR_S "icons");*/
}

// -- crash safe

#if HAVE_UNISTD_H && HAVE_STDLIB_H
RETSIGTYPE signal_crash_handler(int signal)
{
    #ifdef MIDORI_CRASHSAFE
    // TODO: backup session
    #endif
    g_print("Terminated: %s\n", g_strsignal(signal));
    g_print("%s\n", strlen(DEBUG_MSG) ? DEBUG_MSG : "<unknown>");
    exit(1);
}
#endif

// -- main function

int main(int argc, char* *argv)
{
    DEBUG_FUNC
    // TODO: Any way to have a crash handler on non-posix platforms?
    #if HAVE_UNISTD_H && HAVE_STDLIB_H
    signal(SIGSEGV, signal_crash_handler);
    signal(SIGABRT, signal_crash_handler);
    #endif

    // Parse cli options
    gint repeats = 2;
    gboolean bVersion = FALSE;
    GOptionEntry entries[] =
    {
     { "repeats", 'r', 0, G_OPTION_ARG_INT, &repeats, "An unused value", "N" },
     { "version", 'v', 0, G_OPTION_ARG_NONE, &bVersion, "Display program version", NULL }
    };

    GError* error = NULL;
    if(!gtk_init_with_args(&argc, &argv, "[URI]", entries
     , NULL/*GETTEXT_PACKAGE*/, &error))
    {
     g_error_free(error);
     return 1;
    }

    if(bVersion)
    {
     g_print(PACKAGE_STRING " - Copyright (c) 2007 Christian Dywan\n\n"
             "GTK+2:      " GTK_VER "\n"
	     "WebKit:     " WEBKIT_VER "\n"
	     "Libsexy:    " LIBSEXY_VER "\n"
             "GetText:    N/A\n"
	     "\n"
	     "CrashSafe:  " MIDORI_CRASHSAFE_ "\n"
	     "Debugging:  " MIDORI_DEBUG_ "\n"
	     "Engine:     " MIDORI_ENGINE_ "\n"
             "\n"
             "Please report comments, suggestions and bugs to:\n"
             "\t" PACKAGE_BUGREPORT "\n"
             "Check for new versions at:\n"
             "\thttp://software.twotoasts.de\n");
     return 0;
    }

    tooltips = gtk_tooltips_new();

    // Load configuration files
    // TODO: What about default config in a global config folder?
    gchar* sConfig = g_build_filename(g_get_user_config_dir(), PACKAGE_NAME, NULL);
    g_mkdir_with_parents(sConfig, 0755);
    gchar* sConfigFile = g_build_filename(sConfig, "config", NULL);
    config = config_load(sConfigFile, &error); g_free(sConfigFile);
    if(error != NULL)
    {
     g_print("Notice: No configuration loaded. %s", error->message);
     // TODO: This line appears to crash for some people.
     g_error_free(error);
    }
    gchar* sAccelsFile = g_build_filename(sConfig, "accels", NULL);
    gtk_accel_map_load(sAccelsFile); g_free(sAccelsFile);
    /*gchar* sBookmarkFile = g_build_filename(sConfig, "bookmarks.xbel", NULL);
    bookmarks = g_bookmark_file_new();
    if(!g_bookmark_file_load_from_file(bookmarks, sBookmarkFile, &error))
     g_warning("Bookmarks couldn't be loaded. %s", error->message); g_free(sBookmarkFile);*/
    g_free(sConfig);

    //browser_list = g_list_alloc();
    accel_group = gtk_accel_group_new();

    init_stock_icons();
    #ifdef MIDORI_WEBKIT
    webkit_gtk_init();
    #endif

    gchar* sHomepage = argc > 1 ? argv[1] : config->sHomepage;
    // Open as many tabs as we have uris, seperated by pipes
    CBrowser* browser = NULL;
    gchar* sUri = strtok(g_strdup(sHomepage), "|");
    while(sUri != NULL)
    {
     browser = browser_new(browser);
     /* TODO: Use g_idle_add to prevent code from being called
     before the main loop is running */
     document_load_uri(browser->document, magic_uri(sUri));
     sUri = strtok(NULL, "|");
    } g_free(sUri);

    gtk_main();
    DEBUG_FUNC

    // Save configuration files
    sConfig = g_build_filename(g_get_user_config_dir(), PACKAGE_NAME, NULL);
    g_mkdir_with_parents(sConfig, 0755);
    sConfigFile = g_build_filename(sConfig, "config", NULL);
    if(!config_save(config, sConfigFile, &error))
    {
     g_warning("Configuration couldn't be saved. %s", error->message);
     g_error_free(error);
    }
    g_free(sConfigFile);
    sAccelsFile = g_build_filename(sConfig, "accels", NULL);
    gtk_accel_map_save(sAccelsFile); g_free(sAccelsFile);
    /*sBookmarkFile = g_build_filename(sConfig, "bookmarks.xbel", NULL);
    if(!g_bookmark_file_to_file(bookmarks, sBookmarkFile, NULL))
     g_warning("Bookmarks couldn't be saved."); g_free(sBookmarkFile);*/
    g_free(sConfig);
    return 0;
}
